#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import KeyField
from .core import TensorNoInput, TensorLike
from .array import tensor


class TensorEmptyBase(object):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(TensorEmptyBase, self).__init__(*args, **kwargs)
        self._gen_rand()

    def _gen_rand(self):
        if getattr(self, '_rand', None) is None:
            self._obj_set('_rand', np.random.random())

    def to_chunk_op(self, *args):
        op = self.copy().reset_key()
        op._rand = None
        op._gen_rand()
        return op


class TensorEmpty(TensorEmptyBase, TensorNoInput):
    __slots__ = '_rand',
    _op_type_ = OperandDef.TENSOR_EMPTY

    def __init__(self, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype or 'f8')
        super(TensorEmpty, self).__init__(_dtype=dtype, _gpu=gpu, **kw)


def empty(shape, dtype=None, chunk_size=None, gpu=False):
    """
    Return a new tensor of given shape and type, without initializing entries.

    Parameters
    ----------
    shape : int or tuple of int
        Shape of the empty tensor
    dtype : data-type, optional
        Desired output data-type.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        Tensor of uninitialized (arbitrary) data of the given shape, dtype, and
        order.  Object arrays will be initialized to None.

    See Also
    --------
    empty_like, zeros, ones

    Notes
    -----
    `empty`, unlike `zeros`, does not set the array values to zero,
    and may therefore be marginally faster.  On the other hand, it requires
    the user to manually set all the values in the array, and should be
    used with caution.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.empty([2, 2]).execute()
    array([[ -9.74499359e+001,   6.69583040e-309],
           [  2.13182611e-314,   3.06959433e-309]])         #random

    >>> mt.empty([2, 2], dtype=int).execute()
    array([[-1073741821, -1067949133],
           [  496041986,    19249760]])                     #random
    """
    op = TensorEmpty(dtype=dtype, gpu=gpu)
    return op(shape, chunk_size=chunk_size)


class TensorEmptyLike(TensorEmptyBase, TensorLike):
    __slots__ = '_rand',
    _op_type_ = OperandDef.TENSOR_EMPTY_LIKE

    _input = KeyField('input')

    def __init__(self, dtype=None, gpu=None, sparse=False, **kw):
        dtype = np.dtype(dtype) if dtype is not None else None
        super(TensorEmptyLike, self).__init__(_dtype=dtype, _gpu=gpu,
                                              _sparse=sparse, **kw)


def empty_like(a, dtype=None, gpu=None):
    """
    Return a new tensor with the same shape and type as a given tensor.

    Parameters
    ----------
    a : array_like
        The shape and data-type of `a` define these same attributes of the
        returned tensor.
    dtype : data-type, optional
        Overrides the data type of the result.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        Array of uninitialized (arbitrary) data with the same
        shape and type as `a`.

    See Also
    --------
    ones_like : Return a tensor of ones with shape and type of input.
    zeros_like : Return a tensor of zeros with shape and type of input.
    empty : Return a new uninitialized tensor.
    ones : Return a new tensor setting values to one.
    zeros : Return a new tensor setting values to zero.

    Notes
    -----
    This function does *not* initialize the returned tensor; to do that use
    `zeros_like` or `ones_like` instead.  It may be marginally faster than
    the functions that do set the array values.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = ([1,2,3], [4,5,6])                         # a is array-like
    >>> mt.empty_like(a).execute()
    array([[-1073741821, -1073741821,           3],    #ranm
           [          0,           0, -1073741821]])
    >>> a = mt.array([[1., 2., 3.],[4.,5.,6.]])
    >>> mt.empty_like(a).execute()
    array([[ -2.00000715e+000,   1.48219694e-323,  -2.00000572e+000],#random
           [  4.38791518e-305,  -2.00000715e+000,   4.17269252e-309]])
    """
    a = tensor(a)
    op = TensorEmptyLike(dtype=dtype, gpu=gpu, sparse=a.issparse())
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ....compat import izip
from ....operands import BroadcastTo
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


class TensorBroadcastTo(BroadcastTo, TensorOperandMixin):
    def __init__(self, shape=None, dtype=None, sparse=False, **kw):
        super(TensorBroadcastTo, self).__init__(_shape=shape, _dtype=dtype, _sparse=sparse, **kw)

    def _set_inputs(self, inputs):
        super(TensorBroadcastTo, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, tensor, shape):
        return self.new_tensor([tensor], shape)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        in_tensor = op.inputs[0]
        shape = op.shape
        new_dim = tensor.ndim - in_tensor.ndim

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_shape = shape[:new_dim] + tuple(s if in_tensor.shape[idx] != 1 else shape[new_dim+idx]
                                                  for idx, s in enumerate(c.shape))
            chunk_idx = (0,) * new_dim + c.index
            chunk_op = op.copy().reset_key()
            chunk_op._shape = chunk_shape
            out_chunk = chunk_op.new_chunk([c], chunk_shape, index=chunk_idx)
            out_chunks.append(out_chunk)

        nsplits = [tuple(c.shape[i] for c in out_chunks if all(idx == 0 for j, idx in enumerate(c.index) if j != i))
                   for i in range(len(out_chunks[0].shape))]
        new_op = op.copy()
        return new_op.new_tensors([in_tensor], tensor.shape, chunks=out_chunks, nsplits=nsplits)


def broadcast_to(tensor, shape):
    """Broadcast an tensor to a new shape.

    Parameters
    ----------
    tensor : array_like
        The tensor to broadcast.
    shape : tuple
        The shape of the desired array.

    Returns
    -------
    broadcast : Tensor

    Raises
    ------
    ValueError
        If the tensor is not compatible with the new shape according to Mars's
        broadcasting rules.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([1, 2, 3])
    >>> mt.broadcast_to(x, (3, 3)).execute()
    array([[1, 2, 3],
           [1, 2, 3],
           [1, 2, 3]])
    """
    from ...core import Tensor

    tensor = tensor if isinstance(tensor, Tensor) else astensor(tensor)
    shape = tuple(shape) if isinstance(shape, (list, tuple)) else (shape,)

    if tensor.shape == shape:
        return tensor

    new_ndim = len(shape) - tensor.ndim
    if new_ndim < 0:
        raise ValueError('input operand has more dimensions than allowed by the axis remapping')
    if any(o != n for o, n in izip(tensor.shape, shape[new_ndim:]) if o != 1):
        raise ValueError('operands could not be broadcast together '
                         'with remapped shapes [original->remapped]: {0} '
                         'and requested shape {1}'.format(tensor.shape, shape))

    op = TensorBroadcastTo(shape, dtype=tensor.dtype, sparse=tensor.issparse())
    return op(tensor, shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import KeyField, Int32Field
from .core import TensorHasInput
from .zeros import TensorZeros
from .array import tensor


class TensorTri(TensorHasInput):
    def __call__(self, m):
        return self.new_tensor([m], shape=m.shape)

    def to_chunk_op(self, *args):
        k, = args
        op = self.copy().reset_key()
        op._k = k
        return op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        m = op.input
        k = op.k
        is_triu = type(op) == TensorTriu

        fx = lambda x, y: x - y + k
        nsplits = m.nsplits
        cum_size = [np.cumsum(s) for s in nsplits]

        out_chunks = []
        for out_idx in itertools.product(*[range(len(s)) for s in nsplits]):
            i, j = out_idx[-2:]
            ld_pos = cum_size[-2][i] - 1, cum_size[-1][j] - nsplits[-1][j]
            ru_pos = cum_size[-2][i] - nsplits[-2][i], cum_size[-1][j] - 1

            ld_fx = fx(*ld_pos)
            ru_fx = fx(*ru_pos)

            chunk_shape = tuple(nsplits[i][idx] for i, idx in enumerate(out_idx))
            if (is_triu and ld_fx > 0 and ru_fx > 0) or (not is_triu and ld_fx < 0 and ru_fx < 0):
                # does not cross, fill with zeros
                chunk_op = TensorZeros(dtype=op.dtype, gpu=op.gpu, sparse=op.sparse)
                out_chunk = chunk_op.new_chunk(None, chunk_shape, index=out_idx)
            else:
                lu_pos = ru_pos[0], ld_pos[1]
                chunk_k = fx(*lu_pos)

                input_chunk = m.cix[out_idx]
                chunk_op = op.to_chunk_op(chunk_k)
                out_chunk = chunk_op.new_chunk([input_chunk], chunk_shape, index=out_idx)

            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape, chunks=out_chunks, nsplits=m.nsplits)


class TensorTriu(TensorTri):
    _op_type_ = OperandDef.TENSOR_TRIU

    _input = KeyField('input')
    _k = Int32Field('k')

    def __init__(self, k=None, dtype=None, sparse=False, **kw):
        super(TensorTriu, self).__init__(_k=k, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def k(self):
        return self._k


def triu(m, k=0):
    """
    Upper triangle of a tensor.

    Return a copy of a matrix with the elements below the `k`-th diagonal
    zeroed.

    Please refer to the documentation for `tril` for further details.

    See Also
    --------
    tril : lower triangle of a tensor

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.triu([[1,2,3],[4,5,6],[7,8,9],[10,11,12]], -1).execute()
    array([[ 1,  2,  3],
           [ 4,  5,  6],
           [ 0,  8,  9],
           [ 0,  0, 12]])

    """
    m = tensor(m)
    op = TensorTriu(k, dtype=m.dtype, sparse=m.issparse())
    return op(m)


class TensorTril(TensorTri):
    _op_type_ = OperandDef.TENSOR_TRIL

    _input = KeyField('input')
    _k = Int32Field('k')

    def __init__(self, k=None, dtype=None, sparse=False, **kw):
        super(TensorTril, self).__init__(_k=k, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def k(self):
        return self._k


def tril(m, k=0):
    """
    Lower triangle of a tensor.

    Return a copy of a tensor with elements above the `k`-th diagonal zeroed.

    Parameters
    ----------
    m : array_like, shape (M, N)
        Input tensor.
    k : int, optional
        Diagonal above which to zero elements.  `k = 0` (the default) is the
        main diagonal, `k < 0` is below it and `k > 0` is above.

    Returns
    -------
    tril : Tensor, shape (M, N)
        Lower triangle of `m`, of same shape and data-type as `m`.

    See Also
    --------
    triu : same thing, only for the upper triangle

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.tril([[1,2,3],[4,5,6],[7,8,9],[10,11,12]], -1).execute()
    array([[ 0,  0,  0],
           [ 4,  0,  0],
           [ 7,  8,  0],
           [10, 11, 12]])

    """
    m = tensor(m)
    op = TensorTril(k, dtype=m.dtype, sparse=m.issparse())
    return op(m)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import hashlib

from ..utils import to_binary, calc_data_size
from ..compat import functools32
from ..errors import StoreFull, StoreKeyExists


logger = logging.getLogger(__name__)


class PlasmaChunkStore(object):
    """
    Wrapper of plasma client for Mars objects
    """
    def __init__(self, plasma_client):
        from ..serialize.dataserializer import mars_serialize_context

        self._plasma_client = plasma_client
        self._actual_size = None
        self._serialize_context = mars_serialize_context()

    def get_actual_capacity(self):
        """
        Get actual capacity of plasma store
        :return: actual storage size in bytes
        """
        if self._actual_size is None:
            from pyarrow import plasma, lib

            bufs = []
            total_size = 0
            allocate_unit = 4 * 1024 * 1024
            try:
                while True:
                    obj_id = plasma.ObjectID.from_random()
                    bufs.append(self._plasma_client.create(obj_id, allocate_unit))
                    self._plasma_client.seal(obj_id)
                    total_size += allocate_unit
            except lib.PlasmaStoreFull:
                pass
            del bufs
            self._plasma_client.evict(total_size)
            self._actual_size = total_size
        return self._actual_size

    @staticmethod
    @functools32.lru_cache(100)
    def _calc_object_id(session_id, chunk_key):
        """
        Calc unique object id for chunks
        """
        from pyarrow.plasma import ObjectID
        key = '%s#%s' % (session_id, chunk_key)
        digest = hashlib.md5(to_binary(key)).digest()
        return ObjectID(digest + digest[:4])

    def create(self, session_id, chunk_key, size):
        from pyarrow.lib import PlasmaStoreFull, PlasmaObjectExists

        obj_id = self._calc_object_id(session_id, chunk_key)
        try:
            buffer = self._plasma_client.create(obj_id, size)
            return buffer
        except PlasmaStoreFull:
            exc_type = PlasmaStoreFull
            logger.warning('Chunk %s(%d) failed to store to plasma due to StoreFullError',
                           chunk_key, size)
        except PlasmaObjectExists:
            exc_type = PlasmaObjectExists
            logger.warning('Chunk %s(%d) already exists in plasma store', chunk_key, size)

        if exc_type is PlasmaStoreFull:
            raise StoreFull
        elif exc_type is PlasmaObjectExists:
            raise StoreKeyExists

    def seal(self, session_id, chunk_key):
        from pyarrow.lib import PlasmaObjectNonexistent
        obj_id = self._calc_object_id(session_id, chunk_key)
        try:
            self._plasma_client.seal(obj_id)
        except PlasmaObjectNonexistent:
            raise KeyError('(%r, %r)' % (session_id, chunk_key))

    def get(self, session_id, chunk_key):
        """
        Get deserialized Mars object from plasma store
        """
        from pyarrow.plasma import ObjectNotAvailable

        obj_id = self._calc_object_id(session_id, chunk_key)
        obj = self._plasma_client.get(obj_id, serialization_context=self._serialize_context, timeout_ms=10)
        if obj is ObjectNotAvailable:
            raise KeyError('(%r, %r)' % (session_id, chunk_key))
        return obj

    def get_buffer(self, session_id, chunk_key):
        """
        Get raw buffer from plasma store
        """
        obj_id = self._calc_object_id(session_id, chunk_key)
        [buf] = self._plasma_client.get_buffers([obj_id], timeout_ms=10)
        if buf is None:
            raise KeyError('(%r, %r)' % (session_id, chunk_key))
        return buf

    def get_actual_size(self, session_id, chunk_key):
        """
        Get actual size of Mars object from plasma store
        """
        buf = None
        try:
            obj_id = self._calc_object_id(session_id, chunk_key)
            [buf] = self._plasma_client.get_buffers([obj_id], timeout_ms=10)
            size = buf.size
        finally:
            del buf
        return size

    def put(self, session_id, chunk_key, value):
        """
        Put a Mars object into plasma store
        :param session_id: session id
        :param chunk_key: chunk key
        :param value: Mars object to be put
        """
        import pyarrow
        from pyarrow.lib import PlasmaStoreFull, PlasmaObjectExists
        from ..serialize.dataserializer import DataTuple

        data_size = calc_data_size(value)
        if isinstance(value, tuple):
            value = DataTuple(*value)

        obj_id = self._calc_object_id(session_id, chunk_key)

        try:
            serialized = pyarrow.serialize(value, self._serialize_context)
            try:
                buffer = self._plasma_client.create(obj_id, serialized.total_bytes)
                stream = pyarrow.FixedSizeBufferWriter(buffer)
                stream.set_memcopy_threads(6)
                serialized.write_to(stream)
                self._plasma_client.seal(obj_id)
            except PlasmaObjectExists:
                [buffer] = self._plasma_client.get_buffers([obj_id])
            finally:
                del serialized
            return buffer
        except PlasmaStoreFull:
            logger.warning('Chunk %s(%d) failed to store to plasma due to StoreFullError',
                           chunk_key, data_size)
            exc = PlasmaStoreFull
        if exc is PlasmaStoreFull:
            raise StoreFull

    def contains(self, session_id, chunk_key):
        """
        Check if given chunk key exists in current plasma store
        """
        try:
            obj_id = self._calc_object_id(session_id, chunk_key)
            return self._plasma_client.contains(obj_id)
        except KeyError:
            return False



# -*- coding: utf-8 -*-

from bisect import insort
from collections import Counter
from hashlib import md5
from sys import version_info


class KetamaRing(object):
    """Implement a ketama compatible consistent hashing ring."""

    def __init__(self):
        """Create a new HashRing.
        """
        self._distribution = Counter()
        self._keys = []
        self._nodes = {}
        self._replicas = 4
        self._ring = {}

        if version_info >= (3, ):
            self._listbytes = lambda x: x

    def hashi(self, key, replica=0):
        """Returns a ketama compatible hash from the given key.
        """
        dh = self._listbytes(md5(str(key).encode('utf-8')).digest())
        rd = replica * 4
        return (
            (dh[3 + rd] << 24) | (dh[2 + rd] << 16) |
            (dh[1 + rd] << 8) | dh[0 + rd])

    def _hashi_weight_generator(self, node_name, node_conf):
        """Calculate the weight factor of the given node and
        yield its hash key for every configured replica.

        :param node_name: the node name.
        """
        ks = (node_conf['vnodes'] * len(self._nodes) *
              node_conf['weight']) // self._weight_sum
        for w in range(0, ks):
            w_node_name = '%s-%s' % (node_name, w)
            for i in range(0, self._replicas):
                yield self.hashi(w_node_name, replica=i)

    @staticmethod
    def _listbytes(data):
        """Python 2 compatible int iterator from str.

        :param data: the string to int iterate upon.
        """
        return map(ord, data)

    def _create_ring(self, nodes):
        """Generate a ketama compatible continuum/ring.
        """
        _weight_sum = 0
        for node_conf in self._nodes.values():
            _weight_sum += node_conf['weight']
        self._weight_sum = _weight_sum

        _distribution = Counter()
        _keys = []
        _ring = {}
        for node_name, node_conf in self._nodes.items():
            for h in self._hashi_weight_generator(node_name, node_conf):
                _ring[h] = node_name
                insort(_keys, h)
                _distribution[node_name] += 1
        self._distribution = _distribution
        self._keys = _keys
        self._ring = _ring

    def _remove_node(self, node_name):
        """Remove the given node from the continuum/ring.

        :param node_name: the node name.
        """
        try:
            self._nodes.pop(node_name)
        except Exception:
            raise KeyError('node \'{}\' not found, available nodes: {}'.format(
                node_name, self._nodes.keys()))
        else:
            self._create_ring(self._nodes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorUniform(operands.Uniform, TensorRandomOperandMixin):
    __slots__ = '_low', '_high', '_size'
    _input_fields_ = ['_low', '_high']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorUniform, self).__init__(_size=size, _state=state, _dtype=dtype,
                                            _gpu=gpu, **kw)

    def __call__(self, low, high, chunk_size=None):
        return self.new_tensor([low, high], None, raw_chunk_size=chunk_size)


def uniform(random_state, low=0.0, high=1.0, size=None, chunk_size=None, gpu=None, **kw):
    r"""
    Draw samples from a uniform distribution.

    Samples are uniformly distributed over the half-open interval
    ``[low, high)`` (includes low, but excludes high).  In other words,
    any value within the given interval is equally likely to be drawn
    by `uniform`.

    Parameters
    ----------
    low : float or array_like of floats, optional
        Lower boundary of the output interval.  All values generated will be
        greater than or equal to low.  The default value is 0.
    high : float or array_like of floats
        Upper boundary of the output interval.  All values generated will be
        less than high.  The default value is 1.0.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``low`` and ``high`` are both scalars.
        Otherwise, ``mt.broadcast(low, high).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized uniform distribution.

    See Also
    --------
    randint : Discrete uniform distribution, yielding integers.
    random_integers : Discrete uniform distribution over the closed
                      interval ``[low, high]``.
    random_sample : Floats uniformly distributed over ``[0, 1)``.
    random : Alias for `random_sample`.
    rand : Convenience function that accepts dimensions as input, e.g.,
           ``rand(2,2)`` would generate a 2-by-2 array of floats,
           uniformly distributed over ``[0, 1)``.

    Notes
    -----
    The probability density function of the uniform distribution is

    .. math:: p(x) = \frac{1}{b - a}

    anywhere within the interval ``[a, b)``, and zero elsewhere.

    When ``high`` == ``low``, values of ``low`` will be returned.
    If ``high`` < ``low``, the results are officially undefined
    and may eventually raise an error, i.e. do not rely on this
    function to behave when passed arguments satisfying that
    inequality condition.

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> s = mt.random.uniform(-1,0,1000)

    All values are within the given interval:

    >>> mt.all(s >= -1).execute()
    True
    >>> mt.all(s < 0).execute()
    True

    Display the histogram of the samples, along with the
    probability density function:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(s.execute(), 15, normed=True)
    >>> plt.plot(bins, mt.ones_like(bins).execute(), linewidth=2, color='r')
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().uniform(
            handle_array(low), handle_array(high), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorUniform(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(low, high, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.'

from collections import Iterable

import numpy as np

from ....compat import six
from ..utils import validate_axis
from ..datasource import tensor as astensor
from .ravel import ravel


def roll(a, shift, axis=None):
    """
    Roll tensor elements along a given axis.

    Elements that roll beyond the last position are re-introduced at
    the first.

    Parameters
    ----------
    a : array_like
        Input tensor.
    shift : int or tuple of ints
        The number of places by which elements are shifted.  If a tuple,
        then `axis` must be a tuple of the same size, and each of the
        given axes is shifted by the corresponding number.  If an int
        while `axis` is a tuple of ints, then the same value is used for
        all given axes.
    axis : int or tuple of ints, optional
        Axis or axes along which elements are shifted.  By default, the
        tensor is flattened before shifting, after which the original
        shape is restored.

    Returns
    -------
    res : Tensor
        Output tensor, with the same shape as `a`.

    See Also
    --------
    rollaxis : Roll the specified axis backwards, until it lies in a
               given position.

    Notes
    -----

    Supports rolling over multiple dimensions simultaneously.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(10)
    >>> mt.roll(x, 2).execute()
    array([8, 9, 0, 1, 2, 3, 4, 5, 6, 7])

    >>> x2 = mt.reshape(x, (2,5))
    >>> x2.execute()
    array([[0, 1, 2, 3, 4],
           [5, 6, 7, 8, 9]])
    >>> mt.roll(x2, 1).execute()
    array([[9, 0, 1, 2, 3],
           [4, 5, 6, 7, 8]])
    >>> mt.roll(x2, 1, axis=0).execute()
    array([[5, 6, 7, 8, 9],
           [0, 1, 2, 3, 4]])
    >>> mt.roll(x2, 1, axis=1).execute()
    array([[4, 0, 1, 2, 3],
           [9, 5, 6, 7, 8]])

    """
    from ..merge import concatenate

    a = astensor(a)
    raw = a

    if axis is None:
        a = ravel(a)
        axis = 0

    if not isinstance(shift, Iterable):
        shift = (shift,)
    else:
        shift = tuple(shift)
    if not isinstance(axis, Iterable):
        axis = (axis,)
    else:
        axis = tuple(axis)

    [validate_axis(a.ndim, ax) for ax in axis]
    broadcasted = np.broadcast(shift, axis)
    if broadcasted.ndim > 1:
        raise ValueError(
            "'shift' and 'axis' should be scalars or 1D sequences")

    shifts = {ax: 0 for ax in range(a.ndim)}
    for s, ax in broadcasted:
        shifts[ax] += s

    for ax, s in six.iteritems(shifts):
        if s == 0:
            continue

        s = -s
        s %= a.shape[ax]

        slc1 = (slice(None),) * ax + (slice(s, None),)
        slc2 = (slice(None),) * ax + (slice(s),)

        a = concatenate([a[slc1], a[slc2]], axis=ax)

    return a.reshape(raw.shape)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
import os
import time
from collections import defaultdict

from .kvstore import KVStoreActor
from .utils import SchedulerActor
from ..config import options


logger = logging.getLogger(__name__)


class ResourceActor(SchedulerActor):
    """
    Actor managing free resources on workers
    """
    def __init__(self):
        super(ResourceActor, self).__init__()
        self._meta_cache = dict()
        self._kv_store_ref = None
        self._worker_allocations = dict()

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        super(ResourceActor, self).post_create()
        self.ref().clean_worker()

        self._kv_store_ref = self.ctx.actor_ref(KVStoreActor.default_name())
        if not self.ctx.has_actor(self._kv_store_ref):
            self._kv_store_ref = None

    @classmethod
    def default_name(cls):
        return 's:%s' % cls.__name__

    def clean_worker(self):
        """
        Remove worker when it does not update its status for a long time
        """
        for worker in list(self._meta_cache.keys()):
            worker_meta = self._meta_cache[worker]
            if 'update_time' not in worker_meta:
                continue

            if time.time() - worker_meta['update_time'] > options.scheduler.status_timeout:
                del self._meta_cache[worker]

        self.ref().clean_worker(_tell=True, _delay=1)

    def get_worker_count(self):
        return len(self._meta_cache)

    def get_workers_meta(self):
        return copy.deepcopy(self._meta_cache)

    def set_worker_meta(self, worker, worker_meta):
        self._meta_cache[worker] = worker_meta
        if self._kv_store_ref is not None:
            self._kv_store_ref.write('/workers/meta/%s' % worker, json.dumps(worker_meta),
                                     _tell=True, _wait=False)
            self._kv_store_ref.write('/workers/meta_timestamp', str(int(time.time())),
                                     _tell=True, _wait=False)

    def allocate_resource(self, session_id, op_key, endpoint, alloc_dict):
        """
        Try allocate resource for operands
        :param session_id: session id
        :param op_key: operand key
        :param endpoint: worker endpoint
        :param alloc_dict: allocation dict, listing resources needed by the operand
        :return: True if allocated successfully
        """
        worker_stats = self._meta_cache[endpoint]['hardware']
        if endpoint not in self._worker_allocations:
            self._worker_allocations[endpoint] = dict()
        if session_id not in self._worker_allocations[endpoint]:
            self._worker_allocations[endpoint][session_id] = dict()
        worker_allocs = self._worker_allocations[endpoint][session_id]

        res_used = defaultdict(lambda: 0)
        free_alloc_keys = []
        for alloc_op_key, alloc in worker_allocs.items():
            d, t = alloc
            for k, v in d.items():
                res_used[k] += v
        for k in free_alloc_keys:
            self.deallocate_resource(session_id, k, endpoint)

        for k, v in alloc_dict.items():
            res_used[k] += v

        sufficient = True
        for k, v in res_used.items():
            if res_used[k] > worker_stats.get(k, 0):
                sufficient = False
                break

        if sufficient:
            self._worker_allocations[endpoint][session_id][op_key] = (alloc_dict, time.time())
            return True
        else:
            return False

    def deallocate_resource(self, session_id, op_key, endpoint):
        """
        Deallocate resource
        :param session_id: session id
        :param op_key: operand key
        :param endpoint: worker endpoint
        """
        try:
            del self._worker_allocations[endpoint][session_id][op_key]
        except KeyError:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import operator
from collections import Iterable

from ...compat import six, lrange, lmap
from ...operands import Slice
from .array import as_same_device, device


def _get_index(chunk):
    try:
        return chunk.index
    except AttributeError:
        if isinstance(chunk.op, Slice):
            return chunk.inputs[0].index
        raise


def _norm_axis(axis):
    if isinstance(axis, six.integer_types):
        return axis, True
    if isinstance(axis, Iterable):
        axis = sorted(tuple(axis))
        if len(axis) == 1:
            return axis[0], True
        return axis, False

    assert axis is None
    return None, False


def _concatenate(ctx, chunk):
    inputs = [ctx[input.key] for input in chunk.inputs]

    if isinstance(inputs[0], tuple):
        ctx[chunk.key] = tuple(_base_concatenate(chunk, [input[i] for input in inputs])
                               for i in range(len(inputs[0])))
    else:
        ctx[chunk.key] = _base_concatenate(chunk, inputs)


def _base_concatenate(chunk, inputs):
    inputs, device_id, xp = as_same_device(inputs, device=chunk.op.device, ret_extra=True)

    axis, single_axis = _norm_axis(chunk.op.axis)
    if single_axis:
        with device(device_id):
            res = xp.concatenate(tuple(inputs), axis=axis)
    else:
        axes = axis or lrange(chunk.ndim)
        chunks = [(_get_index(input), data) for input, data in zip(chunk.inputs, inputs)]
        with device(device_id):
            for i in range(len(axes) - 1):
                new_chunks = []
                for idx, cs in itertools.groupby(chunks, key=lambda t: t[0][:-1]):
                    cs = lmap(operator.itemgetter(1), cs)
                    new_chunks.append((idx, xp.concatenate(cs, axis=len(axes)-i-1)))
                chunks = new_chunks
            res = xp.concatenate(lmap(operator.itemgetter(1), chunks), axis=axes[0])
    return res


def _stack(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    axis = chunk.op.axis
    with device(device_id):
        ctx[chunk.key] = xp.stack(inputs, axis=axis)


def register_merge_handler():
    from ...operands import Concatenate, Stack
    from .core import register

    register(Concatenate, _concatenate)
    register(Stack, _stack)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from . import kvstore
from .actors import FunctionActor
from .promise import PromiseActor
from .lib.uhashring import HashRing


SCHEDULER_PATH = '/schedulers'
logger = logging.getLogger(__name__)


def create_hash_ring(schedulers):
    return HashRing(nodes=schedulers, hash_fn='ketama')


def get_scheduler(hash_ring, key, size=1):
    if size == 1:
        return hash_ring.get_node(key)

    return tuple(it['nodename'] for it in hash_ring.range(key, size=size))


class _ClusterInfoWatchActor(FunctionActor):
    def __init__(self, service_discover_addr, cluster_info_ref):
        self._client = kvstore.get(service_discover_addr)
        self._cluster_info_ref = cluster_info_ref

        if isinstance(self._client, kvstore.LocalKVStore):
            raise ValueError('etcd_addr should not be a local address, got {0}'.format(service_discover_addr))

    def _get_schedulers(self):
        schedulers = [s.key.rsplit('/', 1)[1] for s in self._client.read(SCHEDULER_PATH).children]
        logger.debug('Schedulers obtained. Results: %r', schedulers)
        return schedulers

    def get_schedulers(self):
        try:
            return self._get_schedulers()
        except KeyError:
            self._client.watch(SCHEDULER_PATH)
            return self._get_schedulers()

    def watch(self):
        for new_schedulers in self._client.eternal_watch(SCHEDULER_PATH):
            self._cluster_info_ref.set_schedulers(new_schedulers)


class ClusterInfoActor(FunctionActor):
    def __init__(self, schedulers=None, service_discover_addr=None):
        if (schedulers is None and service_discover_addr is None) or \
                (schedulers is not None and service_discover_addr is not None):
            raise ValueError('Either schedulers or etcd_addr should be provided')

        self._schedulers = schedulers
        self._hash_ring = None
        self._service_discover_addr = service_discover_addr
        self._watcher = None
        self._observer_refs = []

    @classmethod
    def default_name(cls):
        return cls.__name__

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        if self._service_discover_addr:
            watcher = self._watcher = self.ctx.create_actor(
                _ClusterInfoWatchActor, self._service_discover_addr, self.ref())
            self._schedulers = watcher.get_schedulers()
            watcher.watch(_tell=True)

        self._hash_ring = create_hash_ring(self._schedulers)

    def pre_destroy(self):
        if self._service_discover_addr:
            self.ctx.destroy_actor(self._watcher)

    def register_observer(self, observer, fun_name):
        self._observer_refs.append((self.ctx.actor_ref(observer), fun_name))

    def get_schedulers(self):
        return self._schedulers

    def set_schedulers(self, schedulers):
        self._schedulers = schedulers
        self._hash_ring = create_hash_ring(self._schedulers)

        for observer_ref, fun_name in self._observer_refs:
            # notify the observers to update the new scheduler list
            getattr(observer_ref, fun_name)(schedulers)

    def get_scheduler(self, key, size=1):
        if len(self._schedulers) == 1 and size == 1:
            return self._schedulers[0]
        return get_scheduler(self._hash_ring, key, size=size)


class HasClusterInfoActor(PromiseActor):
    def __init__(self):
        super(HasClusterInfoActor, self).__init__()

        # the scheduler list
        self._schedulers = None
        self._hash_ring = None

        self._cluster_info_ref = None

    def set_schedulers(self, schedulers):
        self._schedulers = schedulers
        if len(schedulers) > 1:
            self._hash_ring = create_hash_ring(schedulers)

    def set_cluster_info_ref(self, set_schedulers_fun_name=None):
        set_schedulers_fun_name = set_schedulers_fun_name or self.set_schedulers.__name__

        # cluster_info_actor is created when scheduler initialized
        self._cluster_info_ref = self.ctx.actor_ref(ClusterInfoActor.default_name())
        # when some schedulers lost, notification will be received
        self._cluster_info_ref.register_observer(self.ref(), set_schedulers_fun_name)
        self.set_schedulers(self._cluster_info_ref.get_schedulers())

    def get_actor_ref(self, key):
        addr = self.get_scheduler(key)
        return self.ctx.actor_ref(addr, key)

    def get_promise_ref(self, key):
        addr = self.get_scheduler(key)
        return self.promise_ref(addr, key)

    def get_scheduler(self, key, size=1):
        if len(self._schedulers) == 1:
            return self._schedulers[0]

        return get_scheduler(self._hash_ring, key, size=size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from contextlib import contextmanager

import numpy as np
try:
    import cupy as cp
except ImportError:
    cp = None

from ...lib import sparse
from ...lib.sparse.core import issparse, get_dense_module


def get_array_module(x, nosparse=False):
    if issparse(x):
        if nosparse:
            return get_dense_module(x)
        return sparse
    if cp:
        return cp.get_array_module(x)
    return np


def array_module(gpu):
    if gpu:
        if cp is None:
            raise ImportError('Execute on GPU requires for `cupy` library')
        return cp

    return np


def _get(x):
    m = get_array_module(x)

    if m is np:
        return x
    if m is sparse:
        return x if not hasattr(x, 'get') else x.get()
    return x.get()


def move_to_device(x, device_id):
    if hasattr(x, 'device') and x.device.id == device_id:
        return x

    assert device_id >= 0

    if issparse(x) and device_id > 0:
        raise NotImplementedError

    # for dense array, we currently copy from gpu to memory and then copy back to destination device
    # to avoid kernel panic
    with cp.cuda.Device(device_id):
        return cp.asarray(x.get())  # remove `get` to do directly copy


def _most_nbytes_device(device_nbytes):
    device_to_nbytes = defaultdict(lambda: 0)
    for device, nbytes in device_nbytes:
        device_to_nbytes[device] += nbytes
    return max(device_to_nbytes, key=lambda i: device_to_nbytes[i])


def as_same_device(inputs, device=None, ret_extra=False):
    input_tensors = [i for i in inputs if hasattr(i, 'ndim') and i.ndim > 0]  # filter scalar
    has_sparse = any(issparse(i) for i in inputs)

    if device is None:
        try:
            device = _most_nbytes_device(
                (i.device.id if hasattr(i, 'device') else -1, i.nbytes) for i in input_tensors)
        except ValueError:
            device = -1

    if device == -1:
        outputs = [_get(i) for i in inputs]
    else:
        outputs = [move_to_device(i, device) for i in inputs]

    if not ret_extra:
        return outputs

    if has_sparse:
        m = sparse
    else:
        if len(input_tensors) > 0:
            m = get_array_module(input_tensors[0])
        else:
            m = np
    return outputs, device, m


def as_np_array(x):
    xp = get_array_module(x)
    return x if xp == np else x.get()


def is_sparse_module(xp):
    return xp is sparse


@contextmanager
def device(device_id):
    if device_id is None or device_id < 0:
        yield
    else:
        with cp.cuda.Device(device_id):
            yield



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from functools import partial

from .. import promise
from ..compat import OrderedDict3, six, functools32
from ..config import options
from ..utils import parse_memory_limit, log_unhandled, readable_size
from ..errors import *
from .utils import WorkerActor

logger = logging.getLogger(__name__)


@log_unhandled
def ensure_chunk(promise_actor, session_id, chunk_key, move_to_end=False):
    from .dispatcher import DispatchActor

    chunk_holder_ref = promise_actor.promise_ref(ChunkHolderActor.default_name())
    dispatch_ref = promise_actor.promise_ref(DispatchActor.default_name())
    if chunk_holder_ref.is_stored(chunk_key):
        logger.debug('No need to load key %s from spill.', chunk_key)
        if move_to_end:
            chunk_holder_ref.move_to_end(chunk_key)
        return promise.Promise(done=True)

    logger.debug('Try starting loading data %s from spill.', chunk_key)

    @log_unhandled
    def _release_mem(*args, **kwargs):
        accepts = kwargs.pop('accepts', True)
        if accepts:
            logger.debug('Chunk %s loaded into plasma.', chunk_key)
        else:
            logger.error('Chunk %s failed to load in plasma.', chunk_key)
            six.reraise(*args)

    uid = dispatch_ref.get_hash_slot('spill', chunk_key)
    spill_ref = promise_actor.promise_ref(uid)

    logger.debug('Requesting data load for chunk %s in %s.', chunk_key, uid)
    return spill_ref.load(session_id, chunk_key, _timeout=options.worker.prepare_data_timeout, _promise=True) \
        .then(_release_mem, partial(_release_mem, accepts=False))


class ChunkHolderActor(WorkerActor):
    def __init__(self, plasma_limit=0):
        super(ChunkHolderActor, self).__init__()
        self._plasma_limit = plasma_limit
        self._cache_chunk_sessions = dict()

        self._chunk_holder = OrderedDict3()
        self._total_size = 0
        self._total_hold = 0
        self._pinned_counter = dict()
        self._spill_pending_keys = set()

        self._total_spill = 0
        self._min_spill_size = 0
        self._max_spill_size = 0

        self._dispatch_ref = None
        self._status_ref = None

    def post_create(self):
        from .dispatcher import DispatchActor
        from .status import StatusActor

        super(ChunkHolderActor, self).post_create()
        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())

        self._plasma_limit = self._chunk_store.get_actual_capacity()
        logger.info('Detected actual plasma store size: %s', readable_size(self._plasma_limit))
        self._total_size = self._plasma_limit
        parse_num, is_percent = parse_memory_limit(options.worker.min_spill_size)
        self._min_spill_size = int(self._plasma_limit * parse_num if is_percent else parse_num)
        parse_num, is_percent = parse_memory_limit(options.worker.max_spill_size)
        self._max_spill_size = int(self._plasma_limit * parse_num if is_percent else parse_num)

        self._status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if self.ctx.has_actor(self._status_ref):
            self._status_ref.set_cache_allocations(
                dict(hold=self._total_hold, total=self._total_size), _tell=True, _wait=False)
        else:
            self._status_ref = None

    def pre_destroy(self):
        for k in list(self._chunk_holder):
            del self._chunk_holder[k]

    @functools32.lru_cache(100)
    def _get_spill_actor_ref(self, chunk_key):
        if not options.worker.spill_directory:
            return None
        try:
            uid = self._dispatch_ref.get_hash_slot('spill', chunk_key)
            return self.promise_ref(uid)
        except KeyError:
            return None

    @promise.reject_on_exception
    @log_unhandled
    def spill_size(self, size, multiplier=1, callback=None):
        request_size = size

        if request_size < self._min_spill_size:
            request_size = self._min_spill_size
        request_size *= multiplier
        if request_size > self._plasma_limit:
            raise SpillExhausted
        if request_size > self._max_spill_size:
            request_size = self._max_spill_size

        logger.debug('Start spilling %d(x%d) bytes from shared cache.', request_size, multiplier)

        if request_size + self._total_hold > self._plasma_limit:
            acc_free = 0
            free_keys = []
            for k in self._chunk_holder:
                if k in self._pinned_counter or k in self._spill_pending_keys:
                    continue
                acc_free += len(self._chunk_holder[k])
                free_keys.append(k)
                self._spill_pending_keys.add(k)
                if request_size + self._total_hold - acc_free <= self._plasma_limit:
                    break

            if not free_keys:
                logger.warning('Cannot spill further. Rejected. request=%d', request_size)
                raise SpillExhausted

            logger.debug('Decide to spill %d chunks. request=%d', len(free_keys), request_size)

            @log_unhandled
            def _release_spill_allocations(key):
                if key in self._spill_pending_keys:
                    self._spill_pending_keys.remove(key)
                if key in self._pinned_counter:
                    return
                if key not in self._chunk_holder:
                    return

                logger.debug('Removing reference of chunk %s from %s when spilling', key, self.uid)
                chunk_size = 0
                if key in self._chunk_holder:
                    chunk_size = len(self._chunk_holder[key])
                    self._total_hold -= chunk_size
                    del self._chunk_holder[key]

                self._plasma_client.evict(chunk_size)

            @log_unhandled
            def _handle_spill_reject(*exc, **kwargs):
                key = kwargs['chunk_key']
                if key in self._spill_pending_keys:
                    self._spill_pending_keys.remove(key)
                six.reraise(*exc)

            @log_unhandled
            def _spill_key(key):
                if key in self._pinned_counter:
                    return
                if key not in self._cache_chunk_sessions:
                    return
                spill_ref = self._get_spill_actor_ref(key)
                logger.debug('Send spill request for key %s to spill actor %s', key, spill_ref.uid)
                return spill_ref.spill(self._cache_chunk_sessions[key], key, _promise=True) \
                    .then(lambda *_: _release_spill_allocations(key)) \
                    .catch(functools32.partial(_handle_spill_reject, chunk_key=key))

            @log_unhandled
            def _finalize_spill(*_):
                logger.debug('Finish spilling %d chunks.', len(free_keys))
                self._plasma_client.evict(request_size)
                if callback:
                    self.tell_promise(callback)

                if self._status_ref:
                    self._status_ref.set_cache_allocations(
                        dict(hold=self._total_hold, total=self._total_size), _tell=True, _wait=False)

            promise.all_(_spill_key(k) for k in free_keys).then(_finalize_spill) \
                .catch(lambda *exc: self.tell_promise(callback, *exc, **dict(_accept=False)))
        else:
            logger.debug('No need to spill. request=%d', request_size)

            self._plasma_client.evict(request_size)
            if callback:
                self.tell_promise(callback)

    @log_unhandled
    def register_chunk(self, session_id, chunk_key):
        if chunk_key in self._chunk_holder:
            self._total_hold -= len(self._chunk_holder[chunk_key])
            del self._chunk_holder[chunk_key]

        self._chunk_holder[chunk_key] = self._chunk_store.get_buffer(session_id, chunk_key)
        self._chunk_holder.move_to_end(chunk_key)

        self._total_hold += len(self._chunk_holder[chunk_key])
        self._cache_chunk_sessions[chunk_key] = session_id
        logger.debug('Chunk %s registered in %s. total_hold=%d', chunk_key, self.uid, self._total_hold)

        if self._status_ref:
            self._status_ref.set_cache_allocations(
                dict(hold=self._total_hold, total=self._total_size), _tell=True, _wait=False)

    @log_unhandled
    def unregister_chunk(self, session_id, chunk_key):
        _ = session_id  # noqa: F841

        spill_ref = self._get_spill_actor_ref(chunk_key)
        if spill_ref:
            spill_ref.delete(chunk_key, _tell=True)

        if chunk_key in self._cache_chunk_sessions:
            del self._cache_chunk_sessions[chunk_key]

        if chunk_key in self._chunk_holder:
            self._total_hold -= self._chunk_holder[chunk_key].size
            del self._chunk_holder[chunk_key]

        logger.debug('Chunk %s unregistered in %s. total_hold=%d', chunk_key, self.uid, self._total_hold)

        if self._status_ref:
            self._status_ref.set_cache_allocations(
                dict(hold=self._total_hold, total=self._total_size), _tell=True, _wait=False)

    def is_stored(self, chunk_key):
        return chunk_key in self._chunk_holder

    def move_to_end(self, chunk_key):
        self._chunk_holder.move_to_end(chunk_key)

    @log_unhandled
    def pin_chunks(self, graph_key, chunk_keys):
        if isinstance(chunk_keys, six.string_types):
            chunk_keys = (chunk_keys,)
        if any(k in self._spill_pending_keys for k in chunk_keys):
            raise PinChunkFailed
        pinned = []
        for k in chunk_keys:
            if k not in self._chunk_holder:
                continue
            if k not in self._pinned_counter:
                self._pinned_counter[k] = set()
            self._pinned_counter[k].add(graph_key)
            pinned.append(k)
        return pinned

    @log_unhandled
    def unpin_chunks(self, graph_key, chunk_keys):
        if isinstance(chunk_keys, six.string_types):
            chunk_keys = (chunk_keys,)
        for k in chunk_keys:
            if k not in self._pinned_counter:
                continue
            self._pinned_counter[k].difference_update((graph_key,))
            if not self._pinned_counter[k]:
                del self._pinned_counter[k]

    def dump_keys(self):
        return list(self._chunk_holder.keys())



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import struct

from ..compat import six, BytesIO

try:
    import numpy as np
except ImportError:
    np = None
try:
    import scipy.sparse as sps
except ImportError:
    sps = None
try:
    import cupy as cp
except ImportError:
    cp = None

from ..lib.sparse import SparseNDArray

try:
    import pyarrow
except ImportError:
    pyarrow = None


BUFFER_SIZE = 256 * 1024


class DummyCompress(object):
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @staticmethod
    def decompress(data):
        return data


try:
    import lz4.frame
    try:
        lz4.frame._compression.BUFFER_SIZE = BUFFER_SIZE
    except AttributeError:
        pass
    lz4_open = functools.partial(lz4.frame.open, block_size=lz4.frame.BLOCKSIZE_MAX1MB)
    lz4_compress = functools.partial(lz4.frame.compress, block_size=lz4.frame.BLOCKSIZE_MAX1MB)
    lz4_compressobj = lz4.frame.LZ4FrameCompressor
    lz4_decompress = lz4.frame.decompress
    lz4_decompressobj = lz4.frame.LZ4FrameDecompressor
except ImportError:
    lz4_open = None
    lz4_compress, lz4_compressobj = None, None
    lz4_decompress, lz4_decompressobj = None, None

SERIAL_VERSION = 0


DATA_FLAG_DENSE = 0
DATA_FLAG_CSR = 1

COMPRESS_FLAG_NONE = 0
COMPRESS_FLAG_LZ4 = 1


compressors = {
    COMPRESS_FLAG_LZ4: lz4_compress,
}
compressobjs = {
    COMPRESS_FLAG_NONE: DummyCompress,
    COMPRESS_FLAG_LZ4: lz4_compressobj,
}
decompressors = {
    COMPRESS_FLAG_LZ4: lz4_decompress,
}
decompressobjs = {
    COMPRESS_FLAG_NONE: DummyCompress,
    COMPRESS_FLAG_LZ4: lz4_decompressobj,
}
compress_openers = {
    COMPRESS_FLAG_LZ4: lz4_open,
}


def load(file, raw=False):
    # version bytes
    file.read(2)
    # total size
    file.read(8)

    compress = struct.unpack('<H', file.read(2))[0]

    if compress != COMPRESS_FLAG_NONE:
        file = compress_openers[compress](file, 'rb')

    try:
        buf = file.read()
    finally:
        if compress != COMPRESS_FLAG_NONE:
            file.close()

    if raw:
        return buf
    else:
        return pyarrow.deserialize(memoryview(buf), mars_serialize_context())


class CompressBufferReader(object):
    def __init__(self, buf, compress):
        self._total_bytes = len(buf)
        self._compress_method = compress
        self._compressor = compressobjs[compress]()
        self._pos = 0
        self._mv = memoryview(buf)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def read(self, byte_num):
        if self._pos == self._total_bytes:
            return b''
        bio = BytesIO()
        if self._pos == 0:
            bio.write(struct.pack('<H', SERIAL_VERSION)
                      + struct.pack('<Q', self._total_bytes)
                      + struct.pack('<H', self._compress_method))
            bio.write(self._compressor.begin())
        while self._pos < self._total_bytes and bio.tell() < byte_num:
            end_pos = min(self._pos + byte_num, self._total_bytes)
            bio.write(self._compressor.compress(self._mv[self._pos:end_pos]))
            if end_pos == self._total_bytes:
                bio.write(self._compressor.flush())
            self._pos = end_pos
        return bio.getvalue()

    def close(self):
        self._mv = None
        self._compressor = None


class DecompressBufferWriter(object):
    def __init__(self, buf):
        import pyarrow
        self._buf = buf
        self._writer = pyarrow.FixedSizeBufferWriter(buf)
        self._writer.set_memcopy_threads(6)
        self._decompressor = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def write(self, data):
        mv = memoryview(data)
        if self._decompressor is not None:
            self._writer.write(self._decompressor.decompress(mv))
        else:
            if len(data) < 12:
                raise IOError('Block size too small')
            compress = struct.unpack('<H', mv[10:12])[0]
            self._decompressor = decompressobjs[compress]()
            if len(data) > 12:
                self._writer.write(self._decompressor.decompress(mv[12:]))

    def close(self):
        if self._writer is not None:
            self._writer.close()
        self._writer = None
        self._buf = None
        self._decompressor = None


def loads(buf, raw=False):
    mv = memoryview(buf)
    compress = struct.unpack('<H', mv[10:12])[0]

    if compress == COMPRESS_FLAG_NONE:
        data = buf[12:]
    else:
        data = decompressors[compress](mv[12:])
    if raw:
        return data
    else:
        return pyarrow.deserialize(memoryview(data), mars_serialize_context())


def dump(obj, file, compress=COMPRESS_FLAG_NONE, raw=False):
    if raw:
        serialized = obj
        data_size = len(serialized)
    else:
        serialized = pyarrow.serialize(obj, mars_serialize_context())
        data_size = serialized.total_bytes

    # version bytes
    file.write(struct.pack('<H', SERIAL_VERSION))
    file.write(struct.pack('<Q', data_size))
    file.write(struct.pack('<H', compress))

    if compress != COMPRESS_FLAG_NONE:
        file = compress_openers[compress](file, 'wb')

    try:
        if raw:
            file.write(serialized)
        else:
            serialized.write_to(file)
    finally:
        if compress != COMPRESS_FLAG_NONE:
            file.close()
    return


def dumps(obj, compress=COMPRESS_FLAG_NONE, raw=False):
    sio = six.BytesIO()
    dump(obj, sio, compress=compress, raw=raw)
    return sio.getvalue()


def peek_serialized_size(file):
    try:
        file.seek(2)
        return struct.unpack('<Q', file.read(8))[0]
    finally:
        file.seek(0)


class DataTuple(tuple):
    pass


def _serialize_numpy_array_list(obj):
    if obj.dtype.str != '|O':
        # Make the array c_contiguous if necessary so that we can call change
        # the view.
        if not obj.flags.c_contiguous:
            obj = np.ascontiguousarray(obj)
        return obj.view('uint8'), obj.dtype.str
    else:
        return obj.tolist(), obj.dtype.str


def _deserialize_numpy_array_list(data):
    if data[1] != '|O':
        assert data[0].dtype == np.uint8
        return data[0].view(data[1])
    else:
        return np.array(data[0], dtype=np.dtype(data[1]))


def _serialize_sparse_csr_list(obj):
    sizes = []
    outputs = [None, None]
    for item in (obj.data, obj.indices, obj.indptr):
        serialized_items = _serialize_numpy_array_list(item)
        outputs.extend(serialized_items)
        sizes.append(len(serialized_items))
    outputs[0] = struct.pack('<' + 'B' * len(sizes), *sizes)
    outputs[1] = struct.pack('<' + 'L' * len(obj.shape), *obj.shape)
    return tuple(outputs)


def _deserialize_sparse_csr_list(data):
    sizes = struct.unpack('<' + 'B' * len(data[0]), data[0])
    shape = struct.unpack('<' + 'L' * (len(data[1]) // 4), data[1])
    data_parts = []
    pos = 2
    for size in sizes:
        data_parts.append(_deserialize_numpy_array_list(data[pos:pos + size]))
        pos += size

    empty_arr = np.zeros(0, dtype=data_parts[0].dtype)
    target_csr = sps.coo_matrix((empty_arr, (empty_arr,) * 2), shape=shape,
                                dtype=data_parts[0].dtype).tocsr()
    target_csr.data, target_csr.indices, target_csr.indptr = data_parts
    return SparseNDArray(target_csr)


def _serialize_data_tuple(obj):
    data_meta = []
    outputs = [None]
    for item in obj:
        if isinstance(item, np.ndarray):
            serialized_items = _serialize_numpy_array_list(item)
            outputs.extend(serialized_items)
            data_meta.append(DATA_FLAG_DENSE)
            data_meta.append(len(serialized_items))
        else:
            serialized_items = _serialize_sparse_csr_list(item)
            outputs.extend(serialized_items)
            data_meta.append(DATA_FLAG_CSR)
            data_meta.append(len(serialized_items))
    outputs[0] = struct.pack('<' + 'B' * len(data_meta), *data_meta)
    return tuple(outputs)


def _deserialize_data_tuple(data):
    data_meta = struct.unpack('<' + 'B' * len(data[0]), data[0])
    pos = 1
    data_parts = []
    for flag, size in zip(data_meta[0::2], data_meta[1::2]):
        if flag == DATA_FLAG_DENSE:
            data_parts.append(_deserialize_numpy_array_list(data[pos:pos + size]))
        elif flag == DATA_FLAG_CSR:
            data_parts.append(_deserialize_sparse_csr_list(data[pos:pos + size]))
        pos += size
    return DataTuple(data_parts)


_serialize_context = None


def mars_serialize_context():
    global _serialize_context
    if _serialize_context is None:
        ctx = pyarrow.default_serialization_context()
        ctx.register_type(SparseNDArray, 'mars.SparseNDArray',
                          custom_serializer=_serialize_sparse_csr_list,
                          custom_deserializer=_deserialize_sparse_csr_list)
        ctx.register_type(DataTuple, 'mars.DataTuple',
                          custom_serializer=_serialize_data_tuple,
                          custom_deserializer=_deserialize_data_tuple)
        _serialize_context = ctx
    return _serialize_context



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import KeyField
from .core import TensorNoInput, TensorLike
from .array import tensor


class TensorOnes(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_ONES

    def __init__(self, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype or 'f8')
        super(TensorOnes, self).__init__(_dtype=dtype, _gpu=gpu, **kw)


def ones(shape, dtype=None, chunk_size=None, gpu=False):
    """
    Return a new tensor of given shape and type, filled with ones.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the new tensor, e.g., ``(2, 3)`` or ``2``.
    dtype : data-type, optional
        The desired data-type for the tensor, e.g., `mt.int8`.  Default is
        `mt.float64`.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        Tensor of ones with the given shape, dtype, and order.

    See Also
    --------
    zeros, ones_like

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.ones(5).execute()
    array([ 1.,  1.,  1.,  1.,  1.])

    >>> mt.ones((5,), dtype=int).execute()
    array([1, 1, 1, 1, 1])

    >>> mt.ones((2, 1)).execute()
    array([[ 1.],
           [ 1.]])

    >>> s = (2,2)
    >>> mt.ones(s).execute()
    array([[ 1.,  1.],
           [ 1.,  1.]])

    """
    op = TensorOnes(dtype=dtype, gpu=gpu)
    return op(shape, chunk_size=chunk_size)


class TensorOnesLike(TensorLike):
    _op_type_ = OperandDef.TENSOR_ONES_LIKE

    _input = KeyField('input')

    def __init__(self, dtype=None, gpu=None, sparse=False, **kw):
        dtype = np.dtype(dtype) if dtype is not None else None
        super(TensorOnesLike, self).__init__(_dtype=dtype, _gpu=gpu, _sparse=sparse, **kw)


def ones_like(a, dtype=None, gpu=None):
    """
    Return a tensor of ones with the same shape and type as a given tensor.

    Parameters
    ----------
    a : array_like
        The shape and data-type of `a` define these same attributes of
        the returned tensor.
    dtype : data-type, optional
        Overrides the data type of the result.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        Tensor of ones with the same shape and type as `a`.

    See Also
    --------
    zeros_like : Return a tensor of zeros with shape and type of input.
    empty_like : Return a empty tensor with shape and type of input.
    zeros : Return a new tensor setting values to zero.
    ones : Return a new tensor setting values to one.
    empty : Return a new uninitialized tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(6)
    >>> x = x.reshape((2, 3))
    >>> x.execute()
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> mt.ones_like(x).execute()
    array([[1, 1, 1],
           [1, 1, 1]])

    >>> y = mt.arange(3, dtype=float)
    >>> y.execute()
    array([ 0.,  1.,  2.])
    >>> mt.ones_like(y).execute()
    array([ 1.,  1.,  1.])

    """
    a = tensor(a)
    op = TensorOnesLike(dtype=dtype, gpu=gpu, sparse=a.issparse())
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import logging

from ..compat import six, functools32
from ..utils import to_binary
from ..actors import Distributor

logger = logging.getLogger(__name__)


class SchedulerDistributor(Distributor):
    @functools32.lru_cache(100)
    def distribute(self, uid):
        if not isinstance(uid, six.string_types) or self.n_process == 1:
            return 0
        id_parts = uid.split(':')
        if id_parts[0] == 's':
            # get process id by hashing uid
            allocate_id = int(hashlib.md5(to_binary(uid)).hexdigest(), 16) % (self.n_process - 1) + 1
            return allocate_id
        else:
            return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Operand
from .base import HasInput
from .. import opcodes as OperandDef
from ..serialize import ValueType, KeyField, ListField, StringField


class Choose(Operand):
    _op_type_ = OperandDef.CHOOSE

    _a = KeyField('a')
    _choices = ListField('choices', ValueType.key)
    _mode = StringField('mode')

    def __setattr__(self, key, value):
        if key == '_mode' and value not in ('raise', 'wrap', 'clip'):
            raise ValueError('mode should be raise, wrap or clip')

        super(Choose, self).__setattr__(key, value)

    @property
    def a(self):
        return self._a

    @property
    def choices(self):
        return self._choices

    @property
    def mode(self):
        return self._mode


class Nonzero(HasInput):
    _op_type_ = OperandDef.NONZERO

    _input = KeyField('input')

    @property
    def output_limit(self):
        return float('inf')



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import random
import sys
import time
from functools import partial
from collections import defaultdict, OrderedDict

from .. import promise
from ..config import options
from ..compat import six
from ..errors import *
from ..utils import deserialize_graph, log_unhandled
from .chunkholder import ensure_chunk
from .spill import spill_exists
from .utils import WorkerActor

logger = logging.getLogger(__name__)

_WORKER_RETRY_ERRORS = (PinChunkFailed, SpillExhausted, ObjectNotInPlasma)


class ExecutionActor(WorkerActor):
    """
    Actor for execution control
    """
    _graph_stages = dict()
    _mem_requests = dict()
    _pin_requests = dict()
    _est_finish_times = dict()
    _retry_delays = defaultdict(lambda: 0)
    _last_dump_time = time.time()
    _stop_requests = set()

    def __init__(self):
        super(ExecutionActor, self).__init__()
        self._chunk_holder_ref = None
        self._dispatch_ref = None
        self._mem_quota_ref = None
        self._status_ref = None

        self._scheduler_resource_ref = None

        self._callbacks = defaultdict(list)
        self._callback_cache = OrderedDict()
        self._size_cache = dict()

    def post_create(self):
        from ..scheduler.resource import ResourceActor
        from .chunkholder import ChunkHolderActor
        from .dispatcher import DispatchActor
        from .quota import MemQuotaActor
        from .status import StatusActor

        super(ExecutionActor, self).post_create()
        self._chunk_holder_ref = self.promise_ref(ChunkHolderActor.default_name())

        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._dispatch_ref.register_free_slot(self.uid, 'execution')

        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())

        scheduler_addr = self.get_scheduler(self.uid)
        self._scheduler_resource_ref = self.ctx.actor_ref(ResourceActor.default_name(),
                                                          address=scheduler_addr)

        self._status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

        self.periodical_dump()

    def periodical_dump(self):
        """
        Periodically dump debug information
        """
        if logger.getEffectiveLevel() > logging.DEBUG:
            return
        cls = type(self)
        if cls._last_dump_time < time.time() - 10:
            cls._last_dump_time = time.time()
            if self._graph_stages:
                self._dump_execution_stages()
        self.ref().periodical_dump(_tell=True, _delay=10)

    @staticmethod
    def _build_load_key(graph_key, chunk_key):
        return '%s_load_memory_%s' % (graph_key, chunk_key)

    @promise.reject_on_exception
    @log_unhandled
    def _fetch_remote_data(self, session_id, graph_key, chunk_key, remote_addr, *_, **kwargs):
        """
        Asynchronously send data receiving command to a remote address
        :param session_id: session id
        :param graph_key: graph key
        :param chunk_key: chunk key
        :param remote_addr: remote server containing provided chunk key
        :return: promise object
        """
        from .dispatcher import DispatchActor

        remote_disp_ref = self.promise_ref(uid=DispatchActor.default_name(),
                                           address=remote_addr)
        ensure_cached = kwargs.pop('ensure_cached', True)

        @log_unhandled
        def _finish_fetch(*_):
            self._chunk_holder_ref.pin_chunks(graph_key, chunk_key)
            if self._chunk_holder_ref.is_stored(chunk_key):
                self._mem_quota_ref.release_quota(self._build_load_key(graph_key, chunk_key))

        @log_unhandled
        def _fetch_step(sender_uid):
            sender_ref = self.promise_ref(sender_uid, address=remote_addr)
            logger.debug('Request for chunk %s transferring from %s', chunk_key, remote_addr)
            return sender_ref.send_data(
                session_id, chunk_key, self.address, ensure_cached=ensure_cached,
                timeout=options.worker.prepare_data_timeout, _promise=True
            ).then(_finish_fetch)

        return remote_disp_ref.get_free_slot('sender', _promise=True).then(_fetch_step)

    def estimate_graph_finish_time(self, graph_key, graph, calc_fetch=True, base_time=None):
        """
        Calc predictions for given chunk graph
        """
        if graph_key not in self._graph_stages:
            return

        from ..tensor.expressions.datasource import TensorFetchChunk
        ops = set(type(c.op).__name__ for c in graph if not isinstance(c.op, TensorFetchChunk))
        op_calc_key = ('calc_speed.' + list(ops)[0]) if len(ops) == 1 else None

        stats = defaultdict(lambda: dict(count=0))
        if self._status_ref:
            stats.update(self._status_ref.get_stats(['disk_read_speed', 'disk_write_speed',
                                                     'net_transfer_speed', op_calc_key]))

        if op_calc_key not in stats:
            return None
        if stats[op_calc_key]['count'] < options.optimize.min_stats_count:
            return None
        if abs(stats[op_calc_key]['count']) < 1e-6:
            return None

        input_size = 0
        net_size = 0
        disk_size = 0
        base_time = base_time or time.time()

        if calc_fetch:
            for c in graph:
                if not isinstance(c.op, TensorFetchChunk):
                    break
                input_size += c.nbytes
                if self._chunk_holder_ref.is_stored(c.key):
                    continue
                if spill_exists(c.key):
                    disk_size += c.nbytes
                else:
                    net_size += c.nbytes

            if stats['net_transfer_speed']['count'] >= options.optimize.min_stats_count:
                base_time += net_size * 1.0 / stats['net_transfer_speed']['mean']
            if stats['disk_read_speed']['count'] >= options.optimize.min_stats_count:
                base_time += disk_size * 1.0 / stats['disk_read_speed']['mean']
            else:
                base_time += disk_size * 1.0 / options.optimize.default_disk_io_speed

        est_finish_time = base_time + input_size * 1.0 / stats[op_calc_key]['mean']

        self._est_finish_times[graph_key] = est_finish_time
        self._status_ref.update_stats(dict(
            min_est_finish_time=min(self._est_finish_times.values()),
            max_est_finish_time=max(self._est_finish_times.values()),
        ), _tell=True, _wait=False)

        self.ref().estimate_graph_finish_time(graph_key, graph, _tell=True, _delay=1)

    def _update_stage_info(self, session_id, key, ops, stage):
        self._graph_stages[key] = (stage, time.time())
        if self._status_ref:
            self._status_ref.update_progress(session_id, key, ops, stage, _tell=True, _wait=False)

    @promise.reject_on_exception
    @log_unhandled
    def execute_graph(self, session_id, graph_key, graph_ser, io_meta, data_sizes, send_targets=None,
                      callback=None):
        """
        Submit graph to the worker and control the execution
        :param session_id: session id
        :param graph_key: graph key
        :param graph_ser: serialized executable graph
        :param io_meta: io meta of the chunk
        :param data_sizes: data size of each input chunk, as a dict
        :param send_targets: targets to send results after execution
        :param callback: promise callback
        """
        from ..tensor.expressions.datasource import TensorFetchChunk
        data_sizes = data_sizes or dict()
        graph = deserialize_graph(graph_ser)
        targets = io_meta['chunks']
        chunks_use_once = set(io_meta.get('input_chunks', [])) - set(io_meta.get('shared_input_chunks', []))

        graph_ops = ','.join(type(c.op).__name__ for c in graph if not isinstance(c.op, TensorFetchChunk))
        logger.debug('Worker graph %s(%s) targeting at %r accepted.', graph_key, graph_ops, targets)

        self._update_stage_info(session_id, graph_key, graph_ops, 'allocate_resource')

        # add callbacks to callback store
        if callback is None:
            callback = []
        elif not isinstance(callback, list):
            callback = [callback]
        self._callbacks[graph_key].extend(callback)
        if graph_key in self._callback_cache:
            del self._callback_cache[graph_key]

        unspill_keys = []
        transfer_keys = []
        calc_keys = set()

        alloc_mem_batch = dict()
        alloc_cache_batch = dict()
        input_chunk_keys = dict()

        if self._status_ref:
            self.estimate_graph_finish_time(graph_key, graph)

        # collect potential allocation sizes
        for chunk in graph:
            if not isinstance(chunk.op, TensorFetchChunk) and chunk.key in targets:
                # use estimated size as potential allocation size
                calc_keys.add(chunk.key)
                alloc_mem_batch[chunk.key] = chunk.nbytes * 2
                alloc_cache_batch[chunk.key] = chunk.nbytes
            else:
                # use actual size as potential allocation size
                input_chunk_keys[chunk.key] = data_sizes.get(chunk.key, chunk.nbytes)

        calc_keys = list(calc_keys)

        keys_to_pin = list(input_chunk_keys.keys())
        try:
            self._pin_requests[graph_key] = set(self._chunk_holder_ref.pin_chunks(graph_key, keys_to_pin))
        except PinChunkFailed:
            # cannot pin input chunks: retry later
            callback = self._callbacks[graph_key]
            self._cleanup_graph(session_id, graph_key)

            retry_delay = self._retry_delays[graph_key] + 0.5 + random.random()
            self._retry_delays[graph_key] = min(1 + self._retry_delays[graph_key], 30)
            self.ref().execute_graph(session_id, graph_key, graph_ser, io_meta, data_sizes, send_targets, callback,
                                     _tell=True, _delay=retry_delay)
            return

        load_chunk_sizes = dict((k, v) for k, v in input_chunk_keys.items()
                                if k not in self._pin_requests[graph_key])
        alloc_mem_batch.update((self._build_load_key(graph_key, k), v)
                               for k, v in load_chunk_sizes.items() if k in chunks_use_once)
        self._chunk_holder_ref.spill_size(sum(alloc_cache_batch.values()), _tell=True)

        # build allocation promises
        batch_alloc_promises = []
        if alloc_mem_batch:
            self._mem_requests[graph_key] = list(alloc_mem_batch.keys())
            batch_alloc_promises.append(self._mem_quota_ref.request_batch_quota(alloc_mem_batch, _promise=True))

        @log_unhandled
        def _prepare_inputs(*_):
            if graph_key in self._stop_requests:
                raise ExecutionInterrupted

            logger.debug('Start preparing input data for graph %s', graph_key)
            self._update_stage_info(session_id, graph_key, graph_ops, 'prepare_inputs')
            prepare_promises = []

            handled_keys = set()
            for chunk in graph:
                if chunk.key in handled_keys:
                    continue
                if not isinstance(chunk.op, TensorFetchChunk):
                    continue
                handled_keys.add(chunk.key)

                if self._chunk_holder_ref.is_stored(chunk.key):
                    # data already in plasma: we just pin it
                    pinned_keys = self._chunk_holder_ref.pin_chunks(graph_key, chunk.key)
                    if chunk.key in pinned_keys:
                        self._mem_quota_ref.release_quota(self._build_load_key(graph_key, chunk.key))
                        continue

                if spill_exists(chunk.key):
                    if chunk.key in chunks_use_once:
                        # input only use in current operand, we only need to load it into process memory
                        continue
                    self._mem_quota_ref.release_quota(self._build_load_key(graph_key, chunk.key))
                    load_fun = partial(lambda gk, ck, *_: self._chunk_holder_ref.pin_chunks(gk, ck),
                                       graph_key, chunk.key)
                    unspill_keys.append(chunk.key)
                    prepare_promises.append(ensure_chunk(self, session_id, chunk.key, move_to_end=True) \
                                            .then(load_fun))
                    continue

                # load data from another worker
                worker_results = self.get_meta_ref(session_id, chunk.key) \
                    .get_workers(session_id, chunk.key)
                if worker_results is None:
                    raise DependencyMissing('Dependency %s not met on sending.' % chunk.key)

                worker_priorities = []
                for w in worker_results.children:
                    _, worker_ip = w.key.rsplit('/', 1)
                    # todo sort workers by speed of network and other possible factors
                    worker_priorities.append((worker_ip, (0, )))

                transfer_keys.append(chunk.key)

                # fetch data from other workers, if one fails, try another
                sorted_workers = sorted(worker_priorities, key=lambda pr: pr[1])
                p = self._fetch_remote_data(session_id, graph_key, chunk.key, sorted_workers[0][0],
                                            ensure_cached=chunk.key not in chunks_use_once)
                for wp in sorted_workers[1:]:
                    p = p.catch(functools.partial(self._fetch_remote_data, session_id, graph_key, chunk.key, wp[0],
                                                  ensure_cached=chunk.key not in chunks_use_once))
                prepare_promises.append(p)

            logger.debug('Graph key %s: Targets %r, unspill keys %r, transfer keys %r',
                         graph_key, targets, unspill_keys, transfer_keys)
            return promise.all_(prepare_promises)

        @log_unhandled
        def _wait_free_slot(*_):
            logger.debug('Waiting for free CPU slot for graph %s', graph_key)
            self._update_stage_info(session_id, graph_key, graph_ops, 'fetch_free_slot')
            return self._dispatch_ref.get_free_slot('cpu', _promise=True)

        @log_unhandled
        def _send_calc_request(calc_uid):
            if graph_key in self._stop_requests:
                raise ExecutionInterrupted

            # get allocation for calc, in case that memory exhausts
            target_allocs = dict()
            for chunk in graph:
                if isinstance(chunk.op, TensorFetchChunk):
                    if not self._chunk_holder_ref.is_stored(chunk.key):
                        alloc_key = self._build_load_key(graph_key, chunk.key)
                        if alloc_key in alloc_mem_batch:
                            target_allocs[alloc_key] = alloc_mem_batch[alloc_key]
                elif chunk.key in targets:
                    target_allocs[chunk.key] = alloc_mem_batch[chunk.key]

            logger.debug('Start calculation for graph %s', graph_key)

            self._update_stage_info(session_id, graph_key, graph_ops, 'calculate')
            calc_ref = self.promise_ref(calc_uid)

            self.estimate_graph_finish_time(graph_key, graph, calc_fetch=False)
            # make sure that memory suffices before actually run execution
            return self._mem_quota_ref.request_batch_quota(target_allocs, _promise=True) \
                .then(lambda *_: self._deallocate_scheduler_resource(session_id, graph_key, delay=2)) \
                .then(lambda *_: calc_ref.calc(session_id, graph_ser, targets, _promise=True))

        @log_unhandled
        def _dump_cache(inproc_uid, save_sizes):
            # do some clean up
            self._deallocate_scheduler_resource(session_id, graph_key)
            inproc_ref = self.promise_ref(inproc_uid)

            if graph_key in self._stop_requests:
                inproc_ref.remove_cache(calc_keys, _tell=True)
                raise ExecutionInterrupted

            self._update_stage_info(session_id, graph_key, graph_ops, 'dump_cache')

            logger.debug('Graph %s: Start putting %r into shared cache. Target actor uid %s.',
                         graph_key, calc_keys, inproc_uid)

            self._chunk_holder_ref.unpin_chunks(graph_key, list(set(c.key for c in graph)), _tell=True)
            if logger.getEffectiveLevel() <= logging.DEBUG:
                self._dump_execution_stages()
                # self._cache_ref.dump_cache_status(_tell=True)

            self._size_cache[graph_key] = save_sizes

            if not send_targets:
                # no endpoints to send, dump keys into shared memory and return
                logger.debug('Worker graph %s(%s) finished execution. Dumping %r into plasma...',
                             graph_key, graph_ops, calc_keys)
                return inproc_ref.dump_cache(calc_keys, _promise=True)
            else:
                # dump keys into shared memory and send
                logger.debug('Worker graph %s(%s) finished execution. Dumping %r into plasma '
                             'while actively transferring %r...',
                             graph_key, graph_ops, calc_keys, send_targets)

                return inproc_ref.dump_cache(calc_keys, _promise=True) \
                        .then(_do_active_transfer)

        @log_unhandled
        def _do_active_transfer(*_):
            # transfer the result chunk to expected endpoints
            @log_unhandled
            def _send_chunk(sender_uid, chunk_key, target_addrs):
                sender_ref = self.promise_ref(sender_uid)
                logger.debug('Request for chunk %s sent to %s', chunk_key, target_addrs)
                return sender_ref.send_data(session_id, chunk_key, target_addrs, ensure_cached=False,
                                            timeout=options.worker.prepare_data_timeout, _promise=True)

            if graph_key in self._mem_requests:
                self._mem_quota_ref.release_quotas(self._mem_requests[graph_key], _tell=True)
                del self._mem_requests[graph_key]

            promises = []
            for key, targets in send_targets.items():
                promises.append(self._dispatch_ref.get_free_slot('sender', _promise=True) \
                                .then(partial(_send_chunk, chunk_key=key, target_addrs=targets)) \
                                .catch(lambda *_: None))
            return promise.all_(promises)

        @log_unhandled
        def _handle_rejection(*exc):
            # some error occurred...
            logger.debug('Entering _handle_rejection() for graph %s', graph_key)
            if logger.getEffectiveLevel() <= logging.DEBUG:
                self._dump_execution_stages()
                # self._cache_ref.dump_cache_status(_tell=True)

            if graph_key in self._stop_requests:
                self._stop_requests.remove(graph_key)

            self._mem_quota_ref.cancel_requests(list(alloc_mem_batch.keys()), _tell=True)

            if not issubclass(exc[0], _WORKER_RETRY_ERRORS):
                # exception not retryable: call back to scheduler
                if isinstance(exc[0], ExecutionInterrupted):
                    logger.warning('Execution of graph %s interrupted.', graph_key)
                else:
                    try:
                        six.reraise(*exc)
                    except:
                        logger.exception('Unexpected error occurred in executing %s', graph_key)
                self._invoke_finish_callbacks(session_id, graph_key, *exc, **dict(_accept=False))
                return

            logger.debug('Graph %s rejected from execution because of %s', graph_key, exc[0].__name__)

            cb = self._callbacks[graph_key]
            self._cleanup_graph(session_id, graph_key)

            if issubclass(exc[0], ObjectNotInPlasma):
                retry_delay = 0
            else:
                retry_delay = self._retry_delays[graph_key] + 0.5 + random.random()
                self._retry_delays[graph_key] = min(1 + self._retry_delays[graph_key], 30)

            self.ref().execute_graph(session_id, graph_key, graph_ser, io_meta, data_sizes, send_targets, cb,
                                     _tell=True, _delay=retry_delay)

        promise.all_(batch_alloc_promises).then(_prepare_inputs) \
            .then(_wait_free_slot).then(_send_calc_request) \
            .then(_dump_cache).then(lambda *_: self._invoke_finish_callbacks(session_id, graph_key,
                                                                             self._size_cache.get(graph_key))) \
            .catch(_handle_rejection)

    def _deallocate_scheduler_resource(self, session_id, graph_key, delay=0):
        try:
            self._scheduler_resource_ref.deallocate_resource(
                session_id, graph_key, self.address, _delay=delay, _tell=True, _wait=False)
        except:
            pass

    def _cleanup_graph(self, session_id, graph_key):
        """
        Do clean up after graph is executed
        :param session_id: session id
        :param graph_key: graph key
        """
        logger.debug('Cleaning callbacks for graph %s', graph_key)
        if graph_key in self._callbacks:
            del self._callbacks[graph_key]
        if graph_key in self._graph_stages:
            del self._graph_stages[graph_key]
        if graph_key in self._est_finish_times:
            del self._est_finish_times[graph_key]

        if graph_key in self._mem_requests:
            self._mem_quota_ref.release_quotas(self._mem_requests[graph_key], _tell=True)
            del self._mem_requests[graph_key]

        if graph_key in self._pin_requests:
            self._chunk_holder_ref.unpin_chunks(graph_key, self._pin_requests[graph_key], _tell=True)
            del self._pin_requests[graph_key]

        if self._status_ref:
            self._status_ref.remove_progress(session_id, graph_key, _tell=True, _wait=False)

    @promise.reject_on_exception
    @log_unhandled
    def add_finish_callback(self, session_id, graph_key, callback):
        """
        Register a callback to callback store
        :param session_id: session id
        :param graph_key: graph key
        :param callback: promise call
        """
        logger.debug('Adding callback %r for graph %s', callback, graph_key)
        self._callbacks[graph_key].append(callback)
        if graph_key in self._callback_cache:
            _, args, kwargs = self._callback_cache[graph_key]
            self._invoke_finish_callbacks(session_id, graph_key, *args, **kwargs)

    @log_unhandled
    def stop_execution(self, graph_key):
        """
        Mark graph for stopping
        :param graph_key: graph key
        """
        logger.debug('Receive stop for graph %s', graph_key)
        if graph_key not in self._graph_stages:
            return
        if self._graph_stages[graph_key][0] == 'allocate_resource':
            try:
                raise ExecutionInterrupted
            except:
                exc_info = sys.exc_info()
            if graph_key in self._mem_requests:
                self._mem_quota_ref.cancel_requests(self._mem_requests[graph_key], exc_info, _tell=True)
        self._stop_requests.add(graph_key)

    @log_unhandled
    def _invoke_finish_callbacks(self, session_id, graph_key, *args, **kwargs):
        """
        Call finish callback when execution is done
        :param session_id: session id
        :param graph_key: graph key
        """
        logger.debug('Send finish callback for graph %s into %d targets', graph_key,
                     len(self._callbacks[graph_key]))
        for cb in self._callbacks[graph_key]:
            self.tell_promise(cb, *args, **kwargs)
        self._cleanup_graph(session_id, graph_key)
        if graph_key in self._retry_delays:
            del self._retry_delays[graph_key]

        if graph_key not in self._callback_cache:
            # preserve callback result for several time to allow add_finish_callback()
            # after execution done
            clean_keys = []
            cur_time = time.time()
            last_finish_time = cur_time - options.worker.callback_preserve_time
            self._callback_cache[graph_key] = (cur_time, args, kwargs)
            for k, tp in self._callback_cache.items():
                if tp[0] < last_finish_time:
                    clean_keys.append(k)
                else:
                    break
            for k in clean_keys:
                del self._callback_cache[k]
                if k in self._size_cache:
                    del self._size_cache[k]

    def _dump_execution_stages(self):
        if logger.getEffectiveLevel() <= logging.DEBUG:
            cur_time = time.time()
            stages = dict((k, (cur_time - v[-1], v[0])) for k, v in self._graph_stages.items())
            logger.debug('Executing stages: %r', stages)

    def handle_process_down(self, halt_refs):
        """
        Handle process down event
        :param halt_refs: actor refs in halt processes
        """
        logger.debug('Process halt detected. Trying to reject affected promises %r.',
                     [ref.uid for ref in halt_refs])
        try:
            raise WorkerProcessStopped
        except WorkerProcessStopped:
            exc_info = sys.exc_info()

        for ref in halt_refs:
            self.reject_promise_ref(ref, *exc_info)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .calc import CpuCalcActor
from .chunkholder import ChunkHolderActor, ensure_chunk
from .daemon import WorkerDaemonActor
from .dispatcher import DispatchActor
from .execution import ExecutionActor
from .prochelper import ProcessHelperActor
from .quota import QuotaActor, MemQuotaActor
from .spill import SpillActor
from .status import StatusActor
from .transfer import SenderActor, ReceiverActor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import Tensor
from .array import tensor as astensor
from .diag import diag


def diagflat(v, k=0, sparse=None, gpu=None, chunk_size=None):
    """
    Create a two-dimensional tensor with the flattened input as a diagonal.

    Parameters
    ----------
    v : array_like
        Input data, which is flattened and set as the `k`-th
        diagonal of the output.
    k : int, optional
        Diagonal to set; 0, the default, corresponds to the "main" diagonal,
        a positive (negative) `k` giving the number of the diagonal above
        (below) the main.
    sparse: bool, optional
        Create sparse tensor if True, False as default
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    out : Tensor
        The 2-D output tensor.

    See Also
    --------
    diag : MATLAB work-alike for 1-D and 2-D tensors.
    diagonal : Return specified diagonals.
    trace : Sum along diagonals.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.diagflat([[1,2], [3,4]]).execute()
    array([[1, 0, 0, 0],
           [0, 2, 0, 0],
           [0, 0, 3, 0],
           [0, 0, 0, 4]])

    >>> mt.diagflat([1,2], 1).execute()
    array([[0, 1, 0],
           [0, 0, 2],
           [0, 0, 0]])

    """
    if not isinstance(v, Tensor):
        v = astensor(v).op.data
    return diag(v.flatten(), k=k, sparse=sparse, gpu=gpu, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorGeometric(operands.Geometric, TensorRandomOperandMixin):
    __slots__ = '_p', '_size'
    _input_fields_ = ['_p']

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorGeometric, self).__init__(_state=state, _size=size, _dtype=dtype,
                                              _gpu=gpu, **kw)

    def __call__(self, p, chunk_size=None):
        return self.new_tensor([p], None, raw_chunk_size=chunk_size)


def geometric(random_state, p, size=None, chunk_size=None, gpu=None, **kw):
    """
    Draw samples from the geometric distribution.

    Bernoulli trials are experiments with one of two outcomes:
    success or failure (an example of such an experiment is flipping
    a coin).  The geometric distribution models the number of trials
    that must be run in order to achieve success.  It is therefore
    supported on the positive integers, ``k = 1, 2, ...``.

    The probability mass function of the geometric distribution is

    .. math:: f(k) = (1 - p)^{k - 1} p

    where `p` is the probability of success of an individual trial.

    Parameters
    ----------
    p : float or array_like of floats
        The probability of success of an individual trial.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``p`` is a scalar.  Otherwise,
        ``mt.array(p).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized geometric distribution.

    Examples
    --------
    Draw ten thousand values from the geometric distribution,
    with the probability of an individual success equal to 0.35:

    >>> import mars.tensor as mt

    >>> z = mt.random.geometric(p=0.35, size=10000)

    How many trials succeeded after a single run?

    >>> ((z == 1).sum() / 10000.).execute()
    0.34889999999999999 #random
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().geometric(
            handle_array(p), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorGeometric(state=random_state._state, size=size, gpu=gpu, **kw)
    return op(p, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def array_equal(a1, a2):
    """
    True if two tensors have the same shape and elements, False otherwise.

    Parameters
    ----------
    a1, a2 : array_like
        Input arrays.

    Returns
    -------
    b : bool
        Returns True if the tensors are equal.

    See Also
    --------
    allclose: Returns True if two tensors are element-wise equal within a
              tolerance.
    array_equiv: Returns True if input tensors are shape consistent and all
                 elements equal.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.array_equal([1, 2], [1, 2]).execute()
    True
    >>> mt.array_equal(mt.array([1, 2]), mt.array([1, 2])).execute()
    True
    >>> mt.array_equal([1, 2], [1, 2, 3]).execute()
    False
    >>> mt.array_equal([1, 2], [1, 4]).execute()
    False

    """
    from ..datasource import tensor as astensor
    from ..datasource.scalar import scalar
    from .all import all

    try:
        a1, a2 = astensor(a1), astensor(a2)
    except Exception:
        return scalar(False)

    if a1.shape != a2.shape:
        return scalar(False)
    return all(astensor(a1 == a2))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ....operands import CopyTo
from ....compat import lrange
from ..utils import unify_chunks, broadcast_shape
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor
from .broadcast_to import broadcast_to


class TensorCopyTo(CopyTo, TensorOperandMixin):
    def __init__(self, casting=None, dtype=None, gpu=None, sparse=None, **kw):
        super(TensorCopyTo, self).__init__(_casting=casting, _dtype=dtype,
                                           _gpu=gpu, _sparse=sparse, **kw)

    def check_inputs(self, inputs):
        if not 2 <= len(inputs) <= 3:
            raise ValueError("inputs' length must be 2 or 3")

    def _set_inputs(self, inputs):
        super(TensorCopyTo, self)._set_inputs(inputs)

        self._src = self._inputs[0]
        self._dst = self._inputs[1]
        if len(self._inputs) > 2:
            self._where = self._inputs[2]

    @staticmethod
    def _extract_inputs(inputs):
        if len(inputs) == 2:
            (src, dst), where = inputs, None
        else:
            src, dst, where = inputs
            if where is True:
                where = None
            else:
                where = astensor(where)

        return src, dst, where

    def __call__(self, *inputs):
        from ..core import Tensor

        src, dst, where = self._extract_inputs(inputs)

        if not isinstance(dst, Tensor):
            raise TypeError('dst has to be a Tensor')

        self._dtype = dst.dtype
        self._gpu = dst.op.gpu
        self._sparse = dst.issparse()

        if not np.can_cast(src.dtype, dst.dtype, casting=self.casting):
            raise TypeError('Cannot cast array from {0!r} to {1!r} '
                            'according to the rule {2!s}'.format(
                src.dtype, dst.dtype, self.casting))

        try:
            broadcast_to(src, dst.shape)
        except ValueError:
            raise ValueError('could not broadcast input array '
                             'from shape {0!r} into shape {1!r}'.format(src.shape, dst.shape))
        if where:
            try:
                broadcast_to(where, dst.shape)
            except ValueError:
                raise ValueError('could not broadcast where mask '
                                 'from shape {0!r} into shape {1!r}'.format(src.shape, dst.shape))

        inps = [src, dst]
        if where is not None:
            inps.append(where)
        ret = self.new_tensor(inps, dst.shape)
        dst.data = ret.data

    @classmethod
    def tile(cls, op):
        inputs = unify_chunks(*[(input, lrange(input.ndim)[::-1]) for input in op.inputs])

        chunk_shapes = [t.chunk_shape if hasattr(t, 'chunk_shape') else t
                        for t in inputs]
        out_chunk_shape = broadcast_shape(*chunk_shapes)

        out_chunks = []
        nsplits = [[None] * shape for shape in out_chunk_shape]
        get_index = lambda idx, t: tuple(0 if t.nsplits[i] == (1,) else ix
                                         for i, ix in enumerate(idx))
        for out_idx in itertools.product(*(map(range, out_chunk_shape))):
            in_chunks = [t.cix[get_index(out_idx[-t.ndim:], t)] if t.ndim != 0 else t.chunks[0]
                         for t in inputs]
            out_chunk = op.copy().reset_key().new_chunk(in_chunks, in_chunks[0].shape, index=out_idx)
            out_chunks.append(out_chunk)
            for i, idx, s in zip(itertools.count(0), out_idx, out_chunk.shape):
                nsplits[i][idx] = s

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape, chunks=out_chunks, nsplits=nsplits)


def copyto(dst, src, casting='same_kind', where=True):
    """
    Copies values from one array to another, broadcasting as necessary.

    Raises a TypeError if the `casting` rule is violated, and if
    `where` is provided, it selects which elements to copy.

    Parameters
    ----------
    dst : Tensor
        The tensor into which values are copied.
    src : array_like
        The tensor from which values are copied.
    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
        Controls what kind of data casting may occur when copying.

          * 'no' means the data types should not be cast at all.
          * 'equiv' means only byte-order changes are allowed.
          * 'safe' means only casts which can preserve values are allowed.
          * 'same_kind' means only safe casts or casts within a kind,
            like float64 to float32, are allowed.
          * 'unsafe' means any data conversions may be done.
    where : array_like of bool, optional
        A boolean tensor which is broadcasted to match the dimensions
        of `dst`, and selects elements to copy from `src` to `dst`
        wherever it contains the value True.
    """
    op = TensorCopyTo(casting=casting)
    return op(src, dst, where)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import os
from collections import OrderedDict

from ..actors import ActorNotExist
from ..config import options
from ..promise import PromiseActor
from ..cluster_info import HasClusterInfoActor


logger = logging.getLogger(__name__)


class WorkerActor(HasClusterInfoActor, PromiseActor):
    """
    Base class of all worker actors, providing necessary utils
    """
    @classmethod
    def default_name(cls):
        return 'w:{0}'.format(cls.__name__)

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())
        try:
            self.set_cluster_info_ref()
        except ActorNotExist:
            pass
        self._init_chunk_store()

    def _init_chunk_store(self):
        import pyarrow.plasma as plasma
        from .chunkstore import PlasmaChunkStore
        self._plasma_client = plasma.connect(options.worker.plasma_socket, '', 0)
        self._chunk_store = PlasmaChunkStore(self._plasma_client)

    def get_meta_ref(self, session_id, chunk_key):
        from ..scheduler.chunkmeta import LocalChunkMetaActor
        addr = self.get_scheduler((session_id, chunk_key))
        return self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr)


class ExpMeanHolder(object):
    """
    Collector of statistics of a given series. The value decays by _factor as time elapses.
    """
    def __init__(self, factor=0.8):
        self._factor = factor
        self._count = 0
        self._v_divided = 0
        self._v_divisor = 0
        self._v2_divided = 0

    def put(self, value):
        self._count += 1
        self._v_divided = self._v_divided * self._factor + value
        self._v_divisor = self._v_divisor * self._factor + 1
        self._v2_divided = self._v2_divided * self._factor + value ** 2

    def count(self):
        return self._count

    def mean(self):
        if self._count == 0:
            return 0
        return self._v_divided * 1.0 / self._v_divisor

    def var(self):
        if self._count == 0:
            return 0
        return self._v2_divided * 1.0 / self._v_divisor - self.mean() ** 2

    def std(self):
        return math.sqrt(self.var())


def concat_operand_keys(graph, sep=','):
    from ..tensor.expressions.datasource import TensorFetchChunk
    graph_op_dict = OrderedDict()
    for c in graph:
        if isinstance(c.op, TensorFetchChunk):
            continue
        graph_op_dict[c.op.key] = type(c.op).__name__
    keys = sep.join(graph_op_dict.keys())
    graph_ops = sep.join(graph_op_dict.values())
    return keys, graph_ops



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..datasource import tensor as astensor


def extract(condition, a):
    """
    Return the elements of a tensor that satisfy some condition.

    This is equivalent to ``mt.compress(ravel(condition), ravel(arr))``.  If
    `condition` is boolean ``mt.extract`` is equivalent to ``arr[condition]``.

    Note that `place` does the exact opposite of `extract`.

    Parameters
    ----------
    condition : array_like
        An array whose nonzero or True entries indicate the elements of `arr`
        to extract.
    a : array_like
        Input tensor of the same size as `condition`.

    Returns
    -------
    extract : Tensor
        Rank 1 tensor of values from `arr` where `condition` is True.

    See Also
    --------
    take, put, copyto, compress, place

    Examples
    --------
    >>> import mars.tensor as mt

    >>> arr = mt.arange(12).reshape((3, 4))
    >>> arr.execute()
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11]])
    >>> condition = mt.mod(arr, 3)==0
    >>> condition.execute()
    array([[ True, False, False,  True],
           [False, False,  True, False],
           [False,  True, False, False]])
    >>> mt.extract(condition, arr).execute()
    array([0, 3, 6, 9])


    If `condition` is boolean:

    >>> arr[condition].execute()
    array([0, 3, 6, 9])

    """
    condition = astensor(condition, dtype=bool)
    return a[condition]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..core import CHUNK_TYPE
from .. import random
from .array import array_module


_RANDOM_OP_TO_NP_METHOD = {
    random.TensorRand: 'rand',
    random.TensorRandn: 'randn',
    random.TensorRandint: 'randint',
    random.TensorRandomIntegers: 'random_integers',
    random.TensorRandomSample: 'random_sample',
    random.TensorChoice: 'choice',
    random.TensorBeta: 'beta',
    random.TensorBinomial: 'binomial',
    random.TensorChisquare: 'chisquare',
    random.TensorDirichlet: 'dirichlet',
    random.TensorExponential: 'exponential',
    random.TensorF: 'f',
    random.TensorGamma: 'gamma',
    random.TensorGeometric: 'geometric',
    random.TensorGumbel: 'gumbel',
    random.TensorHypergeometric: 'hypergeometric',
    random.TensorLaplace: 'laplace',
    random.TensorLogistic: 'logistic',
    random.TensorLognormal: 'lognormal',
    random.TensorLogseries: 'logseries',
    random.TensorMultinomial: 'multinomial',
    random.TensorMultivariateNormal: 'multivariate_normal',
    random.TensorNegativeBinomial: 'negative_binomial',
    random.TensorNoncentralChisquare: 'noncentral_chisquare',
    random.TensorNoncentralF: 'noncentral_f',
    random.TensorNormal: 'normal',
    random.TensorPareto: 'pareto',
    random.TensorPoisson: 'poisson',
    random.TensorRandomPower: 'power',
    random.TensorRayleigh: 'rayleigh',
    random.TensorStandardCauchy: 'standard_cauchy',
    random.TensorStandardExponential: 'standard_exponential',
    random.TensorStandardGamma: 'standard_gamma',
    random.TensorStandardNormal: 'standard_normal',
    random.TensorStandardT: 'standard_t',
    random.TensorTriangular: 'triangular',
    random.TensorUniform: 'uniform',
    random.TensorVonmises: 'vonmises',
    random.TensorWald: 'wald',
    random.TensorWeibull: 'weibull',
    random.TensorZipf: 'zipf',
}


def _rand(ctx, chunk):
    xp = array_module(chunk.op.gpu)
    if chunk.op.state:
        rs = chunk.op.state.random_state
    else:
        if xp == np:
            rs = xp.random.RandomState()
        else:
            rs = xp.random
    get_val = lambda x: ctx[x.key] if isinstance(x, CHUNK_TYPE) else x

    method_name = _RANDOM_OP_TO_NP_METHOD[type(chunk.op)]
    try:
        if method_name in ('rand', 'randn'):
            try:
                res = getattr(rs, method_name)(*chunk.op.size, dtype=chunk.op.dtype)
            except TypeError:
                res = getattr(rs, method_name)(*chunk.op.size)
        elif method_name == 'randint':
            try:
                res = rs.randint(get_val(chunk.op.low), get_val(chunk.op.high), size=chunk.op.size,
                                 dtype=chunk.op.dtype)
            except TypeError:
                res = rs.randint(get_val(chunk.op.low), get_val(chunk.op.high), size=chunk.op.size)
        else:
            try:
                res = getattr(rs, method_name)(*(get_val(getattr(chunk.op, arg)) for arg in chunk.op.args),
                                               dtype=chunk.op.dtype)
            except TypeError:
                res = getattr(rs, method_name)(*(get_val(getattr(chunk.op, arg)) for arg in chunk.op.args))
        if hasattr(res, 'dtype') and res.dtype != chunk.op.dtype:
            res = res.astype(chunk.op.dtype)
        if xp != np:
            with xp.cuda.Device(chunk.op.device or 0):
                ctx[chunk.key] = xp.asarray(res)
        else:
            ctx[chunk.key] = res
    except AttributeError:
        if xp != np:
            if not chunk.op.state:
                rs = np.random.RandomState()
            if method_name in ('rand', 'randn'):
                try:
                    res = getattr(rs, method_name)(*chunk.op.size, dtype=chunk.op.dtype)
                except TypeError:
                    res = getattr(rs, method_name)(*chunk.op.size)
            else:
                try:
                    res = getattr(rs, method_name)(*(get_val(getattr(chunk.op, arg)) for arg in chunk.op.args),
                                                   dtype=chunk.op.dtype)
                except TypeError:
                    res = getattr(rs, method_name)(*(get_val(getattr(chunk.op, arg)) for arg in chunk.op.args))
            if res.dtype != chunk.op.dtype:
                res = res.astype(chunk.op.dtype)
            with xp.cuda.Device(chunk.op.device or 0):
                ctx[chunk.key] = xp.asarray(res)
        else:
            raise


def _multivariate_normal(ctx, chunk):
    xp = array_module(chunk.op.gpu)
    if chunk.op.state:
        rs = chunk.op.state.random_state
    else:
        rs = xp.random.RandomState()

    args = []
    for k in chunk.op.args:
        val = getattr(chunk.op, k, None)
        if isinstance(val, CHUNK_TYPE):
            args.append(ctx[val.key])
        else:
            args.append(val)
    mean, cov = args[:2]
    kw = {}
    if args[2] is not None:
        kw['size'] = args[2]
    if args[3] is not None:
        kw['check_valid'] = args[3]
    if args[4] is not None:
        kw['tol'] = args[4]

    try:
        res = rs.multivariate_normal(mean, cov, **kw)
        if xp != np:
            with xp.cuda.Device(chunk.op.device or 0):
                ctx[chunk.key] = xp.asarray(res)
        else:
            ctx[chunk.key] = res
    except AttributeError:
        if xp != np:
            if not chunk.op.state:
                rs = np.random.RandomState()
            res = rs.multivariate_normal(mean, cov, **kw)
            with xp.cuda.Device(chunk.op.device or 0):
                ctx[chunk.key] = xp.asarray(res)
        else:
            raise


def _distribution(ctx, chunk):
    xp = array_module(chunk.op.gpu)
    if chunk.op.state:
        rs = chunk.op.state.random_state
    else:
        rs = xp.random.RandomState()

    args = []
    for k in chunk.op.args:
        val = getattr(chunk.op, k, None)
        if isinstance(val, CHUNK_TYPE):
            args.append(ctx[val.key])
        else:
            args.append(val)

    method_name = _RANDOM_OP_TO_NP_METHOD[type(chunk.op)]
    try:
        res = getattr(rs, method_name)(*args)
        if xp != np:
            with xp.cuda.Device(chunk.op.device or 0):
                ctx[chunk.key] = xp.asarray(res)
        else:
            ctx[chunk.key] = res
    except AttributeError:
        if xp != np:
            if not chunk.op.state:
                rs = np.random.RandomState()
            res = getattr(rs, method_name)(*args)
            with xp.cuda.Device(chunk.op.device or 0):
                ctx[chunk.key] = xp.asarray(res)
        else:
            raise


def _sparse_randint(ctx, chunk):
    from ...lib.sparse import SparseNDArray
    from ...lib.sparse.core import cps, sps

    xp = array_module(chunk.op.gpu)
    if chunk.op.state:
        rs = chunk.op.state.random_state
    else:
        rs = None

    if chunk.ndim > 2:
        raise NotImplementedError

    low = 1 if chunk.op.low == 0 else chunk.op.low

    rs = rs or xp.random
    size = int(np.ceil(np.prod(chunk.shape) * chunk.op.density))
    xps = cps if chunk.op.gpu else sps
    ij = xp.empty((2, size))
    ij[0] = rs.randint(chunk.shape[0], size=size)
    ij[1] = rs.randint(chunk.shape[1], size=size)
    data = rs.randint(low, chunk.op.high, size=size).astype(chunk.op.dtype)
    m = xps.coo_matrix((data, ij), chunk.shape).tocsr()
    m.data[m.data >= chunk.op.high] = chunk.op.high - 1

    # scipy.sparse is too slow, we remove the precise version due to the performance
    # m = sps.random(*chunk.shape, density=chunk.op.density, format='csr')
    # m.data = (rs or xp.random).randint(low, chunk.op.high, size=m.data.size)\
    #     .astype(chunk.op.dtype)

    ctx[chunk.key] = SparseNDArray(m)


def _randint(ctx, chunk):
    if chunk.issparse():
        return _sparse_randint(ctx, chunk)

    return _rand(ctx, chunk)


def register_random_handler():
    from .core import register
    from ...operands import random as random_op

    register(random_op.SimpleRandomData, _rand)
    register(random_op.Distribution, _distribution)
    register(random.TensorMultivariateNormal, _multivariate_normal)
    register(random.TensorRandint, _randint)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..datasource import tensor as astensor
from .concatenate import concatenate


def column_stack(tup):
    """
    Stack 1-D tensors as columns into a 2-D tensor.

    Take a sequence of 1-D tensors and stack them as columns
    to make a single 2-D tensor. 2-D tensors are stacked as-is,
    just like with `hstack`.  1-D tensors are turned into 2-D columns
    first.

    Parameters
    ----------
    tup : sequence of 1-D or 2-D tensors.
        Tensors to stack. All of them must have the same first dimension.

    Returns
    -------
    stacked : 2-D tensor
        The tensor formed by stacking the given tensors.

    See Also
    --------
    stack, hstack, vstack, concatenate

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array((1,2,3))
    >>> b = mt.array((2,3,4))
    >>> mt.column_stack((a,b)).execute()
    array([[1, 2],
           [2, 3],
           [3, 4]])

    """
    from ..datasource import array

    arrays = []
    for a in tup:
        a = astensor(a)
        if a.ndim < 2:
            a = array(a, ndmin=2).T
        arrays.append(a)

    return concatenate(arrays, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .array import as_same_device, device
from ...compat import izip


def _slice(ctx, chunk):
    ctx[chunk.key] = ctx[chunk.inputs[0].key][tuple(chunk.op.slices)]


def _index(ctx, chunk):
    indexes = tuple(ctx[index.key] if hasattr(index, 'key') else index
                    for index in chunk.op.indexes)
    input = ctx[chunk.inputs[0].key]
    ctx[chunk.key] = input[indexes]


def _index_set_value(ctx, chunk):
    indexes = [ctx[index.key] if hasattr(index, 'key') else index
               for index in chunk.op.indexes]
    input = ctx[chunk.inputs[0].key].copy()
    value = ctx[chunk.op.value.key] if hasattr(chunk.op.value, 'key') else chunk.op.value
    if hasattr(input, 'flags') and not input.flags.writeable:
        input.setflags(write=True)
    input[tuple(indexes)] = value
    ctx[chunk.key] = input


def _choose(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)
    a, choices = inputs[0], inputs[1:]

    with device(device_id):
        ctx[chunk.key] = xp.choose(a, choices, mode=chunk.op.mode)


def _unravel_index(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)
    indices = inputs[0]

    with device(device_id):
        outputs = xp.unravel_index(indices, chunk.op.dims)
        for o, output in izip(chunk.op.outputs, outputs):
            ctx[o.key] = output


def register_indexing_handler():
    from ...operands import Slice, Index, IndexSetValue, Choose, UnravelIndex
    from .core import register

    register(Slice, _slice)
    register(Index, _index)
    register(IndexSetValue, _index_set_value)
    register(Choose, _choose)
    register(UnravelIndex, _unravel_index)
